# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the gear box.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.gear_box

.. autosummary::
    :nosignatures:
    :toctree: gear_box/

    at_gear
    cvt
    manual
    mechanical
    planet
"""

import math
import functools
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
from .cvt import dsp as _cvt_model
from .at_gear import dsp as _at_gear
from .mechanical import dsp as _mechanical
from .manual import dsp as _manual
from .planet import dsp as _planet_model
from co2mpas.utils import reject_outliers

dsp = sh.BlueDispatcher(
    name='Gear box model', description='Models the gear box.'
)


@sh.add_function(dsp, outputs=['gear_box_powers_in'])
def calculate_gear_box_powers_in_v1(gear_box_torques_in, gear_box_speeds_in):
    """
    Calculates gear box power [kW].

    :param gear_box_torques_in:
        Torque required vector [N*m].
    :type gear_box_torques_in: numpy.array | float

    :param gear_box_speeds_in:
        Rotating speed of the wheel [RPM].
    :type gear_box_speeds_in: numpy.array | float

    :return:
        Gear box power [kW].
    :rtype: numpy.array | float
    """
    from ..wheels import calculate_wheel_powers
    return calculate_wheel_powers(gear_box_torques_in, gear_box_speeds_in)


@sh.add_function(dsp, outputs=['gear_box_torques_in'])
def calculate_gear_box_torques_in(gear_box_powers_in, gear_box_speeds_in):
    """
    Calculates torque required vector [N*m].

    :param gear_box_powers_in:
        Gear box power [kW].
    :type gear_box_powers_in: numpy.array | float

    :param gear_box_speeds_in:
        Rotating speed of the wheel [RPM].
    :type gear_box_speeds_in: numpy.array | float

    :return:
        Torque required vector [N*m].
    :rtype: numpy.array | float
    """
    from ..wheels import calculate_wheel_torques
    return calculate_wheel_torques(gear_box_powers_in, gear_box_speeds_in)


@sh.add_function(dsp, outputs=['gear_shifts'])
def calculate_gear_shifts(gears):
    """
    Returns when there is a gear shifting [-].

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :return:
        When there is a gear shifting [-].
    :rtype: numpy.array
    """
    return np.ediff1d(gears, to_begin=[0]) != 0


# noinspection PyPep8Naming
@sh.add_function(dsp, outputs=['gear_box_efficiency_constants'])
def get_gear_box_efficiency_constants(has_torque_converter, gear_box_type):
    """
    Returns vehicle gear box efficiency constants (gbp00, gbp10, and gbp01).

    :param has_torque_converter:
        Does the vehicle use torque converter?
    :type has_torque_converter: bool

    :param gear_box_type:
        Gear box type (manual or automatic or cvt).
    :type gear_box_type: str

    :return:
        Vehicle gear box efficiency constants (gbp00, gbp10, and gbp01).
    :rtype: dict
    """
    PARAMS = dfl.functions.get_gear_box_efficiency_constants.PARAMS
    return PARAMS[has_torque_converter and gear_box_type not in ('cvt', 'none')]


def _linear(x, m, q):
    return x * m + q


def _get_par(obj, key, default=None):
    if default is None:
        default = obj

    try:
        return obj.get(key, default)
    except AttributeError:
        return default


@sh.add_function(dsp, outputs=['gear_box_efficiency_parameters_cold_hot'])
def calculate_gear_box_efficiency_parameters_cold_hot(
        gear_box_efficiency_constants, engine_max_torque):
    """
    Calculates the parameters of gear box efficiency model for cold/hot phases.

    :param gear_box_efficiency_constants:
        Vehicle gear box efficiency constants.
    :type gear_box_efficiency_constants: dict

    :param engine_max_torque:
        Engine Max Torque [N*m].
    :type engine_max_torque: float

    :return:
        Parameters of gear box efficiency model for cold/hot phases:

            - 'hot': `gbp00`, `gbp10`, `gbp01`
            - 'cold': `gbp00`, `gbp10`, `gbp01`
    :rtype: dict
    """

    par = {'hot': {}, 'cold': {}}

    for p in ['hot', 'cold']:
        for k, v in gear_box_efficiency_constants.items():
            m = _get_par(_get_par(v, 'm', default=0.0), p)
            q = _get_par(_get_par(v, 'q', default=0.0), p)
            par[p][k] = _linear(engine_max_torque, m, q)

    return par


dsp.add_function(
    function=functools.partial(sh.replicate_value, copy=False),
    inputs=['gear_box_powers_in'],
    outputs=['gear_box_powers_in_hot', 'gear_box_powers_in_cold']
)


@sh.add_function(dsp, outputs=['gear_box_powers_in_hot'])
def calculate_gear_box_powers_in_hot(
        gear_box_efficiency_parameters_cold_hot, gear_box_powers_out,
        gear_box_speeds_out, gear_box_speeds_in, phase='hot'):
    """
    Calculates the gear box powers in for cold/hot phases.

    :param gear_box_efficiency_parameters_cold_hot:
        Parameters of gear box efficiency model for cold/hot phases:

            - 'hot': `gbp00`, `gbp10`, `gbp01`
            - 'cold': `gbp00`, `gbp10`, `gbp01`
    :type gear_box_efficiency_parameters_cold_hot: dict

    :param gear_box_powers_out:
        Gear box power out vector [kW].
    :type gear_box_powers_out: numpy.array

    :param gear_box_speeds_out:
        Wheel speed vector [RPM].
    :type gear_box_speeds_out: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param phase:
        Cold or hot phase.
    :type phase: str

    :return:
        Gear box powers in for cold/hot phases [kW].
    :rtype: numpy.array
    """
    p, c = gear_box_efficiency_parameters_cold_hot[phase], math.pi / 30000.0
    eff, m, q = p['gbp01'], p['gbp10'] * c, p['gbp00'] * c
    po, wo, wi = gear_box_powers_out, gear_box_speeds_out, gear_box_speeds_in
    pi = np.maximum(po + wi * (m * wi + q), 0) / eff
    pi = np.where(po > 0, pi, np.minimum(0, eff * po + wo * (m * wo + q)))
    return np.where(gear_box_speeds_out == gear_box_speeds_in, po, pi)


dsp.add_func(
    functools.partial(calculate_gear_box_powers_in_hot, phase='cold'),
    function_id='calculate_gear_box_powers_in_cold',
    outputs=['gear_box_powers_in_cold']
)

dsp.add_function(
    function=functools.partial(sh.replicate_value, copy=False),
    inputs=['gear_box_powers_out'],
    outputs=['gear_box_powers_out_hot', 'gear_box_powers_out_cold']
)


@sh.add_function(dsp, outputs=['gear_box_powers_out_hot'])
def calculate_gear_box_powers_out_hot(
        gear_box_efficiency_parameters_cold_hot, gear_box_powers_in,
        gear_box_speeds_out, gear_box_speeds_in, phase='hot'):
    """
    Calculates the gear box powers out for cold/hot phases.

    :param gear_box_efficiency_parameters_cold_hot:
        Parameters of gear box efficiency model for cold/hot phases:

            - 'hot': `gbp00`, `gbp10`, `gbp01`
            - 'cold': `gbp00`, `gbp10`, `gbp01`
    :type gear_box_efficiency_parameters_cold_hot: dict

    :param gear_box_powers_in:
        Gear box power in vector [kW].
    :type gear_box_powers_in: numpy.array

    :param gear_box_speeds_out:
        Wheel speed vector [RPM].
    :type gear_box_speeds_out: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param phase:
        Cold or hot phase.
    :type phase: str

    :return:
        Gear box powers out for cold/hot phases [kW].
    :rtype: numpy.array
    """
    p, c = gear_box_efficiency_parameters_cold_hot[phase], math.pi / 30000.0
    eff, m, q = p['gbp01'], p['gbp10'] * c, p['gbp00'] * c
    pi, wo, wi = gear_box_powers_in, gear_box_speeds_out, gear_box_speeds_in
    po = np.minimum(pi - wo * (m * wo + q), 0) / eff
    po = np.where(pi < 0, po, np.maximum(0, eff * pi - wi * (m * wi + q)))
    return np.where(gear_box_speeds_out == gear_box_speeds_in, pi, po)


dsp.add_func(
    functools.partial(calculate_gear_box_powers_out_hot, phase='cold'),
    function_id='calculate_gear_box_powers_out_cold',
    outputs=['gear_box_powers_out_cold']
)

dsp.add_data(
    'has_gear_box_thermal_management',
    dfl.values.has_gear_box_thermal_management
)


@sh.add_function(dsp, outputs=['equivalent_gear_box_heat_capacity'])
def calculate_equivalent_gear_box_heat_capacity(
        engine_mass, has_gear_box_thermal_management):
    """
    Calculates the equivalent gear box heat capacity [kg*J/K].

    :param engine_mass:
        Engine mass [kg].
    :type engine_mass: str

    :param has_gear_box_thermal_management:
        Does the gear box have some additional technology to heat up faster?
    :type has_gear_box_thermal_management: bool

    :return:
       Equivalent gear box heat capacity [kg*J/K].
    :rtype: float
    """

    par = dfl.functions.calculate_engine_heat_capacity.PARAMS

    heated_eng_mass = engine_mass * sum(par['heated_mass_percentage'].values())

    par = dfl.functions.calculate_equivalent_gear_box_heat_capacity
    par = par.PARAMS

    heated_gear_box_mass = heated_eng_mass * par['gear_box_mass_engine_ratio']

    if has_gear_box_thermal_management:
        heated_gear_box_mass *= par['thermal_management_factor']

    return par['heat_capacity']['oil'] * heated_gear_box_mass


dsp.add_data(
    'gear_box_temperature_references',
    dfl.values.gear_box_temperature_references
)


@sh.add_function(dsp, function_id='calculate_gear_box_temperatures', inputs=[
    'gear_box_powers_out', 'gear_box_powers_in_hot', 'gear_box_powers_in_cold',
    'initial_gear_box_temperature', 'engine_thermostat_temperature', 'times',
    'equivalent_gear_box_heat_capacity', 'gear_box_temperature_references'
], outputs=['gear_box_temperatures'])
@sh.add_function(dsp, function_id='calculate_gear_box_temperatures_v1', inputs=[
    'gear_box_powers_in', 'gear_box_powers_out_hot', 'gear_box_powers_out_cold',
    'initial_gear_box_temperature', 'engine_thermostat_temperature', 'times',
    'equivalent_gear_box_heat_capacity', 'gear_box_temperature_references'
], outputs=['gear_box_temperatures'])
def calculate_gear_box_temperatures(
        gear_box_powers, gear_box_powers_hot, gear_box_powers_cold,
        initial_gear_box_temperature, engine_thermostat_temperature, times,
        equivalent_gear_box_heat_capacity, gear_box_temperature_references):
    """
    Calculates the gear box temperatures [°C].

    :param gear_box_powers:
        Gear box power out/in vector [kW].
    :type gear_box_powers: numpy.array

    :param gear_box_powers_hot:
        Gear box powers in/out for hot phase [kW].
    :type gear_box_powers_hot: numpy.array

    :param gear_box_powers_in_cold:
        Gear box powers in/out for cold phase [kW].
    :type gear_box_powers_cold: numpy.array

    :param initial_gear_box_temperature:
        Initial gear box temperature [°C].
    :type initial_gear_box_temperature: float

    :param engine_thermostat_temperature:
        Engine thermostat temperature [°C].
    :type engine_thermostat_temperature: float

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param equivalent_gear_box_heat_capacity:
        Equivalent gear box heat capacity [kg*J/K].
    :type equivalent_gear_box_heat_capacity: float

    :param gear_box_temperature_references:
        Reference temperature [°C].
    :type gear_box_temperature_references: (float, float)

    :return:
        Temperature vector [°C].
    :rtype: numpy.array
    """
    (tc, th), t = gear_box_temperature_references, initial_gear_box_temperature

    temp = np.empty((times.shape[0] + 1,), float)
    temp[0] = t
    tm = temp[1:] = engine_thermostat_temperature - 5.0

    c = 1000.0 / equivalent_gear_box_heat_capacity * np.ediff1d(times, to_end=0)
    ph, pc, p = gear_box_powers_hot, gear_box_powers_cold, gear_box_powers

    it = zip(ph - p, pc - p, (ph - pc) / ((th - tc) or 1), c)
    for i, (dph, dpc, m, c) in enumerate(it, start=1):
        if t >= th or m == 0:
            dp = dph
        elif t <= tc:
            dp = dpc
        else:
            dp = dph + (th - t) * m
        t += abs(dp) * c
        if t >= tm:
            break
        temp[i] = t
    return temp[:-1]


@sh.add_function(dsp, function_id='calculate_gear_box_powers_in', inputs=[
    'gear_box_temperatures', 'gear_box_powers_in_hot',
    'gear_box_powers_in_cold', 'gear_box_temperature_references'
], outputs=['gear_box_powers_in'])
@sh.add_function(dsp, function_id='calculate_gear_box_powers_out', inputs=[
    'gear_box_temperatures', 'gear_box_powers_out_hot',
    'gear_box_powers_out_cold', 'gear_box_temperature_references'
], outputs=['gear_box_powers_out'])
def calculate_gear_box_powers(
        gear_box_temperatures, gear_box_powers_hot, gear_box_powers_cold,
        gear_box_temperature_references):
    """
    Calculates gear box power in/out [kW].

    :param gear_box_powers_hot:
        Gear box powers in/out for hot phase [kW].
    :type gear_box_powers_hot: numpy.array

    :param gear_box_powers_cold:
        Gear box powers in/out for cold phase [kW].
    :type gear_box_powers_cold: numpy.array

    :param gear_box_temperatures:
        Temperature vector [°C].
    :type gear_box_temperatures: numpy.array

    :param gear_box_temperature_references:
        Reference temperature [°C].
    :type gear_box_temperature_references: (float, float)

    :return:
        Gear box power in/out [kW].
    :rtype: numpy.array
    """
    (tc, th), t = gear_box_temperature_references, gear_box_temperatures
    p = gear_box_powers_hot.copy()
    b = (tc < t) & (t < th)
    p[b] += (th - t[b]) / (th - tc) * (p[b] - gear_box_powers_cold[b])
    b = t <= tc
    p[b] = gear_box_powers_cold[b]
    return p


@sh.add_function(dsp, outputs=['gear_box_efficiencies'])
def calculate_gear_box_efficiencies(gear_box_powers_out, gear_box_powers_in):
    """
    Calculates gear box efficiency vector [-].

    :param gear_box_powers_out:
        Gear box power out vector [kW].
    :type gear_box_powers_out: numpy.array

    :param gear_box_powers_in:
        Gear box power in vector [kW].
    :type gear_box_powers_in: numpy.array

    :return:
        Gear box efficiency vector [-].
    :rtype: numpy.array
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        eff = gear_box_powers_in / gear_box_powers_out
    eff[np.isnan(eff) & ~np.isfinite(eff)] = 1
    b = eff > 1
    eff[b] = 1 / eff[b]
    return eff


@sh.add_function(dsp, outputs=['gear_box_mean_efficiency'])
def identify_gear_box_mean_efficiency(gear_box_efficiencies):
    """
    Identify gear box mean efficiency [-].

    :param gear_box_efficiencies:
        Gear box efficiency vector [-].
    :type gear_box_efficiencies: numpy.array

    :return:
        Gear box mean efficiency [-].
    :rtype: float
    """
    return reject_outliers(gear_box_efficiencies)[0]


dsp.add_function(
    function=sh.bypass,
    inputs=['gear_box_mean_efficiency'],
    outputs=['gear_box_mean_efficiency_guess']
)


@sh.add_function(
    dsp, outputs=['gear_box_mean_efficiency_guess'], weight=sh.inf(10, 30)
)
def calculate_gear_box_mean_efficiency_guess(
        times, motive_powers, final_drive_mean_efficiency, gear_box_speeds_in,
        equivalent_gear_box_heat_capacity, gear_box_temperature_references,
        gear_box_efficiency_parameters_cold_hot, initial_gear_box_temperature,
        gear_box_speeds_out):
    """
    Calculate gear box mean efficiency guess [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param final_drive_mean_efficiency:
        Final drive mean efficiency [-].
    :type final_drive_mean_efficiency: float

    :param gear_box_speeds_in:
        Engine speed vector [RPM].
    :type gear_box_speeds_in: numpy.array

    :param equivalent_gear_box_heat_capacity:
        Equivalent gear box heat capacity [kg*J/K].
    :type equivalent_gear_box_heat_capacity: float

    :param gear_box_temperature_references:
        Reference temperature [°C].
    :type gear_box_temperature_references: (float, float)

    :param gear_box_efficiency_parameters_cold_hot:
        Parameters of gear box efficiency model for cold/hot phases:

            - 'hot': `gbp00`, `gbp10`, `gbp01`
            - 'cold': `gbp00`, `gbp10`, `gbp01`
    :type gear_box_efficiency_parameters_cold_hot: dict

    :param initial_gear_box_temperature:
        Initial gear box temperature [°C].
    :type initial_gear_box_temperature: float

    :param gear_box_speeds_out:
        Wheel speed vector [RPM].
    :type gear_box_speeds_out: numpy.array

    :return:
        Gear box mean efficiency guess [-].
    :rtype: float
    """
    from ..electrics.motors.p4 import calculate_motor_p4_electric_powers as func
    p = func(motive_powers, final_drive_mean_efficiency)
    func = functools.partial(
        calculate_gear_box_powers_in_hot,
        gear_box_efficiency_parameters_cold_hot, p, gear_box_speeds_out,
        gear_box_speeds_in
    )
    h, c = func(), func(phase='cold')
    return identify_gear_box_mean_efficiency(calculate_gear_box_efficiencies(
        p, calculate_gear_box_powers(calculate_gear_box_temperatures(
            p, h, c, initial_gear_box_temperature, 100, times,
            equivalent_gear_box_heat_capacity, gear_box_temperature_references
        ), h, c, gear_box_temperature_references)
    ))


# noinspection PyMissingOrEmptyDocstring
def is_automatic(kwargs):
    return kwargs.get('gear_box_type') == 'automatic'


# noinspection PyMissingOrEmptyDocstring
def is_manual(kwargs):
    b = kwargs.get('gear_box_type') == 'manual'
    return b and kwargs.get('cycle_type', 'NEDC') != 'NEDC'


# noinspection PyMissingOrEmptyDocstring
def is_cvt_or_none(kwargs):
    return kwargs.get('gear_box_type') in ('cvt', 'none')


# noinspection PyMissingOrEmptyDocstring
def is_planetary(kwargs):
    return kwargs.get('gear_box_type') == 'planetary'


# noinspection PyMissingOrEmptyDocstring
def is_manual_or_automatic(kwargs):
    return kwargs.get('gear_box_type') in ('manual', 'automatic')


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_mechanical,
    inputs=(
        'accelerations', 'change_gear_window_width', 'engine_max_torque',
        'engine_speed_at_max_velocity', 'engine_speeds_out', 'f0', 'on_engine',
        'final_drive_ratios', 'first_gear_box_ratio', 'full_load_curve',
        'gear_box_ratios', 'gear_box_speeds_out', 'gears', 'idle_engine_speed',
        'last_gear_box_ratio', 'maximum_vehicle_laden_mass', 'maximum_velocity',
        'n_gears', 'plateau_acceleration', 'r_dynamic', 'road_loads',
        'stop_velocity', 'times', 'velocities', 'velocity_speed_ratios',
        'motive_powers', 'correct_gear', 'gear_box_speeds_in',
        {'gear_box_type': sh.SINK}
    ),
    outputs=(
        'engine_speed_at_max_velocity', 'first_gear_box_ratio', 'max_gear',
        'gear_box_ratios', 'gear_box_speeds_in', 'gears', 'last_gear_box_ratio',
        'maximum_velocity', 'speed_velocity_ratios', 'n_gears',
        'velocity_speed_ratios'
    ),
    input_domain=is_manual_or_automatic
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_at_gear,
    dsp_id='at_gear_shifting',
    inputs=(
        'CMV', 'CMV_Cold_Hot', 'DTGS', 'GSPV', 'GSPV_Cold_Hot', 'MVL', 'times',
        'accelerations', 'change_gear_window_width', 'cycle_type', 'gears',
        'engine_temperatures', 'engine_speeds_out', 'full_load_curve',
        'fuel_saving_at_strategy', 'max_velocity_full_load_correction',
        'idle_engine_speed', 'stop_velocity', 'time_cold_hot_transition',
        'motive_powers', 'plateau_acceleration', 'specific_gear_shifting',
        'use_dt_gear_shifting', 'velocities', 'velocity_speed_ratios',
        {'gear_box_type': sh.SINK}
    ),
    outputs=(
        {
            'CMV': ('CMV', 'gear_shifting_model_raw'),
            'CMV_Cold_Hot': ('CMV_Cold_Hot', 'gear_shifting_model_raw'),
            'DTGS': ('DTGS', 'gear_shifting_model_raw'),
            'GSPV': ('GSPV', 'gear_shifting_model_raw'),
            'GSPV_Cold_Hot': ('GSPV_Cold_Hot', 'gear_shifting_model_raw')
        }, 'MVL', 'gears', 'specific_gear_shifting', 'correct_gear'),
    input_domain=is_automatic
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_manual,
    dsp_id='manual_gear_shifting',
    inputs=(
        'cycle_type', 'full_load_speeds', 'idle_engine_speed', 'motive_powers',
        'engine_max_speed', 'full_load_curve', 'engine_max_power', 'road_loads',
        'velocity_speed_ratios', 'engine_speed_at_max_power', 'velocities',
        'accelerations', 'times', {'gear_box_type': sh.SINK}
    ),
    outputs=('gears', 'correct_gear', {
        'MGS': ('MGS', 'gear_shifting_model_raw')
    }),
    input_domain=is_manual
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_cvt_model,
    dsp_id='cvt_model',
    inputs=(
        'CVT', 'accelerations', 'engine_speeds_out', 'gear_box_powers_out',
        'idle_engine_speed', 'on_engine', 'stop_velocity', 'velocities',
        'gear_box_speeds_in', {'gear_box_type': sh.SINK}
    ),
    outputs=(
        'gear_box_speeds_in', 'correct_gear', 'gears', 'max_gear',
        'max_speed_velocity_ratio', {'CVT': ('CVT', 'gear_shifting_model_raw')}
    ),
    input_domain=is_cvt_or_none
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_planet_model,
    dsp_id='no_model',
    inputs=(
        'accelerations', 'engine_speeds_out', 'idle_engine_speed', 'velocities',
        'stop_velocity', 'gear_box_speeds_in', 'gear_box_speeds_out',
        {'gear_box_type': sh.SINK}
    ),
    outputs=(
        'max_speed_velocity_ratio', 'gear_shifting_model', 'gear_box_speeds_in',
        'correct_gear', 'max_gear', 'gears',
    ),
    input_domain=is_planetary
)


@sh.add_function(dsp, outputs=['gear_shifting_model'])
def initialize_gear_shifting_model(
        gear_shifting_model_raw, velocity_speed_ratios, cycle_type):
    """
    Initialize the gear shifting model.

    :param gear_shifting_model_raw:
        A gear shifting model (cmv or gspv or dtgs).
    :type gear_shifting_model_raw: GSPV | CMV | DTGS

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :return:
        Initialized gear shifting model (cmv or gspv or dtgs).
    :rtype: GSPV | CMV | DTGS
    """
    # noinspection PyProtectedMember
    from .at_gear import _upgrade_gsm
    gsm = gear_shifting_model_raw
    return _upgrade_gsm(gsm, velocity_speed_ratios, cycle_type)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the drive battery (high voltage).
"""

import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl

dsp = sh.BlueDispatcher(
    name='Drive Battery',
    description='Models the drive battery (e.g., high voltage).'
)


@sh.add_function(dsp, outputs=['drive_battery_voltages'])
def calculate_drive_battery_voltages(
        drive_battery_electric_powers, drive_battery_currents):
    """
    Calculate the drive battery voltage [V].

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param drive_battery_currents:
        Drive battery current vector [A].
    :type drive_battery_currents: numpy.array

    :return:
        Drive battery voltage [V].
    :rtype: numpy.array
    """
    from .service import calculate_service_battery_currents as func
    return func(drive_battery_electric_powers, drive_battery_currents)


@sh.add_function(dsp, outputs=['drive_battery_currents'])
def calculate_drive_battery_currents(
        drive_battery_electric_powers, drive_battery_voltages):
    """
    Calculate the drive battery current vector [A].

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param drive_battery_voltages:
        Drive battery voltage [V].
    :type drive_battery_voltages: numpy.array

    :return:
        Drive battery current vector [A].
    :rtype: numpy.array
    """
    from .service import calculate_service_battery_currents as func
    return func(drive_battery_electric_powers, drive_battery_voltages)


@sh.add_function(dsp, outputs=['drive_battery_currents'], weight=1)
def calculate_drive_battery_currents_v1(
        drive_battery_capacity, times, drive_battery_state_of_charges):
    """
    Calculate the drive battery current vector [A].

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param drive_battery_state_of_charges:
        State of charge of the drive battery [%].
    :type drive_battery_state_of_charges: numpy.array

    :return:
        Drive battery current vector [A].
    :rtype: numpy.array
    """
    from .service import calculate_service_battery_currents_v1 as func
    return func(drive_battery_capacity, times, drive_battery_state_of_charges)


@sh.add_function(dsp, outputs=['drive_battery_capacity'])
def identify_drive_battery_capacity(
        times, drive_battery_currents, drive_battery_state_of_charges):
    """
    Identify drive battery capacity [Ah].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param drive_battery_currents:
        Drive battery current vector [A].
    :type drive_battery_currents: numpy.array

    :param drive_battery_state_of_charges:
        State of charge of the drive battery [%].
    :type drive_battery_state_of_charges: numpy.array

    :return:
        Drive battery capacity [Ah].
    :rtype: float
    """
    from .service import identify_service_battery_capacity as func
    return func(times, drive_battery_currents, drive_battery_state_of_charges)


@sh.add_function(dsp, outputs=['drive_battery_capacity_kwh'])
def calculate_drive_battery_capacity_kwh(
        drive_battery_capacity, drive_battery_nominal_voltage):
    """
    Calculate drive battery capacity [kWh].

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :param drive_battery_nominal_voltage:
        Drive battery nominal voltage [V].
    :type drive_battery_nominal_voltage: float

    :return:
        Drive battery capacity [kWh].
    :rtype: float
    """
    from .service import calculate_service_battery_capacity_kwh as func
    return func(drive_battery_capacity, drive_battery_nominal_voltage)


@sh.add_function(dsp, outputs=['drive_battery_capacity'])
def calculate_drive_battery_capacity(
        drive_battery_capacity_kwh, drive_battery_nominal_voltage):
    """
    Calculate drive battery capacity [Ah].

    :param drive_battery_capacity_kwh:
        Drive battery capacity [kWh].
    :type drive_battery_capacity_kwh: float

    :param drive_battery_nominal_voltage:
        Drive battery nominal voltage [V].
    :type drive_battery_nominal_voltage: float

    :return:
        Drive battery capacity [Ah].
    :rtype: float
    """
    from .service import calculate_service_battery_capacity as func
    return func(drive_battery_capacity_kwh, drive_battery_nominal_voltage)


@sh.add_function(dsp, outputs=['drive_battery_nominal_voltage'])
def calculate_drive_battery_nominal_voltage(
        drive_battery_capacity_kwh, drive_battery_capacity):
    """
    Calculate drive battery nominal voltage [V].

    :param drive_battery_capacity_kwh:
        Drive battery capacity [kWh].
    :type drive_battery_capacity_kwh: float

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :return:
        Drive battery nominal voltage [V].
    :rtype: float
    """
    from .service import calculate_service_battery_nominal_voltage as func
    return func(drive_battery_capacity_kwh, drive_battery_capacity)


@sh.add_function(dsp, outputs=['drive_battery_electric_powers'])
def calculate_drive_battery_electric_powers(
        drive_battery_currents, drive_battery_voltages):
    """
    Calculate the drive battery electric power [kW].

    :param drive_battery_currents:
        Drive battery current vector [A].
    :type drive_battery_currents: numpy.array

    :param drive_battery_voltages:
        Drive battery voltage [V].
    :type drive_battery_voltages: float

    :return:
        Drive battery electric power [kW].
    :rtype: numpy.array
    """
    from .service import calculate_service_battery_electric_powers as func
    return func(drive_battery_currents, drive_battery_voltages)


@sh.add_function(dsp, outputs=['drive_battery_electric_powers'], weight=1)
def calculate_drive_battery_electric_powers_v1(
        drive_battery_loads, motors_electric_powers,
        dcdc_converter_electric_powers_demand):
    """
    Calculate the drive battery electric power [kW].

    :param drive_battery_loads:
        Drive battery load vector [kW].
    :type drive_battery_loads: numpy.array

    :param motors_electric_powers:
        Cumulative motors electric power [kW].
    :type motors_electric_powers: numpy.array

    :param dcdc_converter_electric_powers_demand:
        DC/DC converter electric power demand [kW].
    :type dcdc_converter_electric_powers_demand: numpy.array

    :return:
        Drive battery electric power [kW].
    :rtype: numpy.array
    """
    p = drive_battery_loads - motors_electric_powers
    p += dcdc_converter_electric_powers_demand
    return p


@sh.add_function(dsp, outputs=['initial_drive_battery_state_of_charge'])
def identify_initial_drive_battery_state_of_charge(
        drive_battery_state_of_charges):
    """
    Identify the initial state of charge of drive battery [%].

    :param drive_battery_state_of_charges:
        State of charge of the drive battery [%].
    :type drive_battery_state_of_charges: numpy.array

    :return:
        Initial state of charge of the drive battery [%].
    :rtype: float
    """
    return drive_battery_state_of_charges[0]


@sh.add_function(
    dsp, outputs=['initial_drive_battery_state_of_charge'], weight=10
)
def default_initial_drive_battery_state_of_charge(
        electrical_hybridization_degree):
    """
    Return the default initial state of charge of drive battery [%].

    :param electrical_hybridization_degree:
        Electrical hybridization degree (i.e., mild, full, plugin, electric).
    :type electrical_hybridization_degree: str

    :return:
        Initial state of charge of the drive battery [%].
    :rtype: float
    """
    d = dfl.functions.default_initial_drive_battery_state_of_charge
    return d.initial_state_of_charge[electrical_hybridization_degree]


@sh.add_function(dsp, outputs=['drive_battery_state_of_charges'])
def calculate_drive_battery_state_of_charges(
        drive_battery_capacity, initial_drive_battery_state_of_charge,
        times, drive_battery_currents):
    """
    Calculates the state of charge of the drive battery [%].

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :param initial_drive_battery_state_of_charge:
        Initial state of charge of the drive battery [%].
    :type initial_drive_battery_state_of_charge: float

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param drive_battery_currents:
        Drive battery current vector [A].
    :type drive_battery_currents: numpy.array

    :return:
        State of charge of the drive battery [%].
    :rtype: numpy.array
    """
    from .service import calculate_service_battery_state_of_charges as func
    return func(
        drive_battery_capacity, initial_drive_battery_state_of_charge, times,
        drive_battery_currents
    )


@sh.add_function(dsp, outputs=['drive_battery_loads'])
def calculate_drive_battery_loads(
        drive_battery_electric_powers, motors_electric_powers,
        dcdc_converter_electric_powers_demand):
    """
    Calculates drive battery load vector [kW].

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param motors_electric_powers:
        Cumulative motors electric power [kW].
    :type motors_electric_powers: numpy.array

    :param dcdc_converter_electric_powers_demand:
        DC/DC converter electric power demand [kW].
    :type dcdc_converter_electric_powers_demand: numpy.array

    :return:
        Drive battery load vector [kW].
    :rtype: numpy.array
    """
    p = drive_battery_electric_powers + motors_electric_powers
    p -= dcdc_converter_electric_powers_demand
    return p


dsp.add_data('drive_battery_load', 0, sh.inf(11, 0))


@sh.add_function(dsp, outputs=['drive_battery_loads'])
def calculate_drive_battery_loads_v1(times, drive_battery_load):
    """
    Calculates drive battery load vector [kW].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param drive_battery_load:
        Drive electric load [kW].
    :type drive_battery_load: float

    :return:
        Drive battery load vector [kW].
    :rtype: numpy.array
    """
    return np.tile(drive_battery_load, times.shape)


@sh.add_function(dsp, outputs=['drive_battery_load'])
def identify_drive_battery_load(drive_battery_loads):
    """
    Identifies drive electric load [kW].

    :param drive_battery_loads:
        Drive battery load vector [kW].
    :type drive_battery_loads: numpy.array

    :return:
        Drive electric load [kW].
    :rtype: float
    """
    return co2_utl.reject_outliers(drive_battery_loads)[0]


@sh.add_function(dsp, outputs=['drive_battery_delta_state_of_charge'])
def calculate_drive_battery_delta_state_of_charge(
        drive_battery_state_of_charges):
    """
    Calculates the overall delta state of charge of the drive battery [%].

    :param drive_battery_state_of_charges:
        State of charge of the drive battery [%].
    :type drive_battery_state_of_charges: numpy.array

    :return:
        Overall delta state of charge of the drive battery [%].
    :rtype: float
    """
    from .service import calculate_service_battery_delta_state_of_charge as func
    return func(drive_battery_state_of_charges)


# noinspection PyMissingOrEmptyDocstring,PyProtectedMember
class DriveBatteryModel:
    def __init__(self, service_battery_model, drive_battery_load,
                 initial_drive_battery_state_of_charge, drive_battery_capacity,
                 dcdc_converter_efficiency=.95, r0=None, ocv=None,
                 n_parallel_cells=1, n_series_cells=1):
        self.r0, self.ocv = r0, ocv
        self.np, self.ns = n_parallel_cells, n_series_cells
        r0 is not None and ocv is not None and self.compile()
        self.service = service_battery_model
        self.dcdc_efficiency = dcdc_converter_efficiency
        self._d_soc = drive_battery_capacity * 36.0 * 2
        self.init_soc = initial_drive_battery_state_of_charge
        self.drive_battery_load = drive_battery_load
        self.reset()
        self._dcdc_p = self.service.nominal_voltage / 1e3

    # noinspection PyAttributeOutsideInit
    def reset(self):
        self.service.reset()
        self._prev_current = 0
        self._prev_soc = self.init_soc

    # noinspection PyAttributeOutsideInit
    def compile(self):
        self._ocv = self.ocv * self.ns / 1e3
        self.constant = self.r0 <= 0

        self.v_b = self.ocv * self.ns / 2
        self.v_b2 = self.v_b ** 2

        if self.constant:
            self._r0 = self.v_ac4 = 0
        else:
            n = self.np / self.r0
            self.c_b, self.c_ac4 = -self.ocv * n / 2, 1e3 / self.ns * n
            self.c_b2 = self.c_b ** 2

            self.v_ac4 = self.ns / n * 1e3

            self._r0 = self.ns / n / 1e3

    def fit(self, currents, voltages):
        from sklearn.linear_model import RANSACRegressor
        m = RANSACRegressor(random_state=0)
        try:
            m.fit(currents[:, None], voltages)
            r0, ocv = float(m.estimator_.coef_), float(m.estimator_.intercept_)
        except ValueError:  # RANSAC failure switch to constant model.
            r0 = ocv = dfl.EPS
        if r0 <= dfl.EPS:
            r0, ocv = 0, np.mean(voltages)
        self.r0 = r0 * self.np / self.ns
        self.ocv = ocv / self.ns
        self.compile()
        return self

    def currents(self, powers):
        if self.constant:
            return powers / self._ocv
        cur = np.sqrt(np.maximum(self.c_b2 + self.c_ac4 * powers, 0)) + self.c_b
        return cur

    def powers(self, currents):
        return (self._ocv + currents * self._r0) * currents

    def voltages(self, powers):
        vol = self.v_b + np.sqrt(np.maximum(self.v_b2 + self.v_ac4 * powers, 0))
        return vol

    def __call__(self, current, time, motive_power, acceleration, on_engine,
                 starter_current=0, prev_soc=None, update=True,
                 service_kw=None):
        dt = time - self.service._prev_time
        from ..motors.p4 import calculate_motor_p4_powers_v1 as f
        dcdc_p = f(self.service(
            time, motive_power, acceleration, on_engine, starter_current,
            update=update, **(service_kw or {})
        )[2] * self._dcdc_p, self.dcdc_efficiency) + self.drive_battery_load

        if prev_soc is None:
            prev_soc = self._prev_soc

        current = current + self.currents(dcdc_p)
        dsoc = (current + self._prev_current) * dt / self._d_soc
        soc = max(0.0, min(prev_soc + dsoc, 100.0))

        if update:
            self._prev_soc, self._prev_current = soc, current

        return soc


@sh.add_function(dsp, outputs=['drive_battery_model'])
def calibrate_drive_battery_model(
        service_battery_model, initial_drive_battery_state_of_charge,
        drive_battery_capacity, drive_battery_n_parallel_cells,
        drive_battery_n_series_cells, drive_battery_currents,
        dcdc_converter_efficiency, drive_battery_load, drive_battery_voltages):
    """
    Calibrate the drive battery current model.

    :param service_battery_model:
         Service battery model.
    :type service_battery_model: ServiceBatteryModel

    :param initial_drive_battery_state_of_charge:
        Initial state of charge of the drive battery [%].
    :type initial_drive_battery_state_of_charge: float

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :param drive_battery_n_parallel_cells:
        Number of battery cells in parallel [-].
    :type drive_battery_n_parallel_cells: int

    :param drive_battery_n_series_cells:
        Number of battery cells in series [-].
    :type drive_battery_n_series_cells: int

    :param drive_battery_currents:
        Drive battery current vector [A].
    :type drive_battery_currents: numpy.array

    :param dcdc_converter_efficiency:
        DC/DC converter efficiency [-].
    :type dcdc_converter_efficiency: float

    :param drive_battery_load:
        Drive electric load [kW].
    :type drive_battery_load: float

    :param drive_battery_voltages:
        Drive battery voltage [V].
    :type drive_battery_voltages: numpy.array

    :return:
        Drive battery current model.
    :rtype: DriveBatteryModel
    """
    return DriveBatteryModel(
        service_battery_model, drive_battery_load,
        initial_drive_battery_state_of_charge, drive_battery_capacity,
        dcdc_converter_efficiency,
        n_parallel_cells=drive_battery_n_parallel_cells,
        n_series_cells=drive_battery_n_series_cells
    ).fit(drive_battery_currents, drive_battery_voltages)


@sh.add_function(dsp, outputs=['drive_battery_r0', 'drive_battery_ocv'])
def get_drive_battery_r0_and_ocv(drive_battery_model):
    """
    Returns drive battery resistance and open circuit voltage [ohm, V].

    :param drive_battery_model:
        Drive battery current model.
    :type drive_battery_model: DriveBatteryModel

    :return:
        Drive battery resistance and open circuit voltage [ohm, V].
    :rtype: float, float
    """
    return drive_battery_model.r0, drive_battery_model.ocv


dsp.add_data('drive_battery_n_cells', 1)


@sh.add_function(dsp, outputs=['drive_battery_nominal_voltage'])
def identify_drive_battery_nominal_voltage(drive_battery_voltages):
    """
    Identify the drive battery nominal voltage [V].

    :param drive_battery_voltages:
        Drive battery voltage [V].
    :type drive_battery_voltages: numpy.array

    :return:
        Drive battery nominal voltage [V].
    :rtype: float
    """
    return np.median(drive_battery_voltages[drive_battery_voltages > dfl.EPS])


@sh.add_function(dsp, outputs=['drive_battery_voltages'], weight=sh.inf(15, 0))
def define_drive_battery_voltages(times, drive_battery_nominal_voltage):
    """
    Defines drive battery voltage vector [V].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param drive_battery_nominal_voltage:
        Drive battery nominal voltage [V].
    :type drive_battery_nominal_voltage: float

    :return:
        Drive battery voltage [V].
    :rtype: numpy.array
    """
    return np.tile(drive_battery_nominal_voltage, times.size)


@sh.add_function(dsp, outputs=['drive_battery_n_parallel_cells'])
def calculate_drive_battery_n_parallel_cells(
        drive_battery_n_cells, drive_battery_n_series_cells):
    """
    Calculate the number of battery cells in parallel [-].

    :param drive_battery_n_cells:
        Number of battery cells [-].
    :type drive_battery_n_cells: int

    :param drive_battery_n_series_cells:
        Number of battery cells in series [-].
    :type drive_battery_n_series_cells: int

    :return:
        Number of battery cells in parallel [-].
    :rtype: int
    """
    return drive_battery_n_cells / drive_battery_n_series_cells


dsp.add_data('drive_battery_technology', dfl.values.drive_battery_technology)


@sh.add_function(dsp, outputs=['drive_battery_n_parallel_cells'], weight=5)
def calculate_drive_battery_n_parallel_cells_v1(
        drive_battery_technology, drive_battery_nominal_voltage,
        drive_battery_n_cells):
    """
    Calculate the number of battery cells in parallel [-].

    :param drive_battery_technology:
        Drive battery technology type (e.g., NiMH, Li-NCA (Li-Ni-Co-Al), etc.).
    :type drive_battery_technology: str

    :param drive_battery_n_cells:
        Number of battery cells [-].
    :type drive_battery_n_cells: int

    :param drive_battery_nominal_voltage:
        Drive battery nominal voltage [V].
    :type drive_battery_nominal_voltage: float

    :return:
        Number of battery cells in parallel [-].
    :rtype: int
    """
    v = dfl.functions.calculate_drive_battery_n_parallel_cells_v1.reference_volt
    v = v.get(drive_battery_technology, v['unknown'])
    n = np.ceil(v / drive_battery_nominal_voltage * drive_battery_n_cells)
    n = int(min(n, drive_battery_n_cells))
    while n < drive_battery_n_cells and drive_battery_n_cells % n:
        n += 1
    return n


@sh.add_function(dsp, outputs=['drive_battery_n_series_cells'])
def calculate_drive_battery_n_series_cells(
        drive_battery_n_cells, drive_battery_n_parallel_cells):
    """
    Calculate the number of battery cells in parallel [-].

    :param drive_battery_n_cells:
        Number of battery cells [-].
    :type drive_battery_n_cells: int

    :param drive_battery_n_parallel_cells:
        Number of battery cells in parallel [-].
    :type drive_battery_n_parallel_cells: int

    :return:
        Number of battery cells in series [-].
    :rtype: int
    """
    return drive_battery_n_cells / drive_battery_n_parallel_cells


@sh.add_function(dsp, outputs=['drive_battery_n_cells'])
def calculate_drive_battery_n_cells(
        drive_battery_n_parallel_cells, drive_battery_n_series_cells):
    """
    Calculate the number of battery cells [-].

    :param drive_battery_n_parallel_cells:
        Number of battery cells in parallel [-].
    :type drive_battery_n_parallel_cells: int

    :param drive_battery_n_series_cells:
        Number of battery cells in series [-].
    :type drive_battery_n_series_cells: int

    :return:
        Number of battery cells [-].
    :rtype: int
    """
    return drive_battery_n_parallel_cells * drive_battery_n_series_cells


@sh.add_function(dsp, outputs=['drive_battery_model'])
def define_drive_battery_model(
        service_battery_model, drive_battery_load, dcdc_converter_efficiency,
        initial_drive_battery_state_of_charge, drive_battery_capacity,
        drive_battery_r0, drive_battery_ocv, drive_battery_n_parallel_cells,
        drive_battery_n_series_cells):
    """
    Define the drive battery current model.

    :param service_battery_model:
         Service battery model.
    :type service_battery_model: ServiceBatteryModel

    :param dcdc_converter_efficiency:
        DC/DC converter efficiency [-].
    :type dcdc_converter_efficiency: float

    :param drive_battery_load:
        Drive electric load [kW].
    :type drive_battery_load: float

    :param initial_drive_battery_state_of_charge:
        Initial state of charge of the drive battery [%].
    :type initial_drive_battery_state_of_charge: float

    :param drive_battery_capacity:
        Drive battery capacity [Ah].
    :type drive_battery_capacity: float

    :param drive_battery_r0:
        Drive battery resistance [ohm].
    :type drive_battery_r0: float

    :param drive_battery_ocv:
        Drive battery open circuit voltage [V].
    :type drive_battery_ocv: float

    :param drive_battery_n_parallel_cells:
        Number of battery cells in parallel [-].
    :type drive_battery_n_parallel_cells: int

    :param drive_battery_n_series_cells:
        Number of battery cells in series [-].
    :type drive_battery_n_series_cells: int

    :return:
        Drive battery current model.
    :rtype: DriveBatteryModel
    """
    return DriveBatteryModel(
        service_battery_model, drive_battery_load,
        initial_drive_battery_state_of_charge, drive_battery_capacity,
        dcdc_converter_efficiency, drive_battery_r0, drive_battery_ocv,
        drive_battery_n_parallel_cells, drive_battery_n_series_cells
    )


@sh.add_function(dsp, outputs=['drive_battery_voltages'])
def calculate_drive_battery_voltages_v1(
        drive_battery_electric_powers, drive_battery_model):
    """
    Calculate the drive battery current vector [A].

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param drive_battery_model:
        Drive battery current model.
    :type drive_battery_model: DriveBatteryModel

    :return:
        Drive battery current vector [A].
    :rtype: numpy.array
    """
    return drive_battery_model.voltages(drive_battery_electric_powers)


@sh.add_function(dsp, outputs=['motors_electric_powers'])
def calculate_motors_electric_powers(
        motor_p0_electric_powers, motor_p1_electric_powers,
        motor_p2_planetary_electric_powers, motor_p2_electric_powers,
        motor_p3_front_electric_powers, motor_p3_rear_electric_powers,
        motor_p4_front_electric_powers, motor_p4_rear_electric_powers):
    """
    Calculate motors electric power [kW].

    :param motor_p0_electric_powers:
        Electric power of motor P0 [kW].
    :type motor_p0_electric_powers: numpy.array | float

    :param motor_p1_electric_powers:
        Electric power of motor P1 [kW].
    :type motor_p1_electric_powers: numpy.array | float

    :param motor_p2_planetary_electric_powers:
        Electric power of planetary motor P2 [kW].
    :type motor_p2_planetary_electric_powers: numpy.array | float

    :param motor_p2_electric_powers:
        Electric power of motor P2 [kW].
    :type motor_p2_electric_powers: numpy.array | float

    :param motor_p3_front_electric_powers:
        Electric power of motor P3 front [kW].
    :type motor_p3_front_electric_powers: numpy.array | float

    :param motor_p3_rear_electric_powers:
        Electric power of motor P3 rear [kW].
    :type motor_p3_rear_electric_powers: numpy.array | float

    :param motor_p4_front_electric_powers:
        Electric power of motor P4 front [kW].
    :type motor_p4_front_electric_powers: numpy.array | float

    :param motor_p4_rear_electric_powers:
        Electric power of motor P4 rear [kW].
    :type motor_p4_rear_electric_powers: numpy.array | float

    :return:
        Cumulative motors electric power [kW].
    :rtype: numpy.array | float
    """
    p = motor_p0_electric_powers + motor_p1_electric_powers
    p += motor_p2_planetary_electric_powers
    p += motor_p2_electric_powers
    p += motor_p3_front_electric_powers
    p += motor_p3_rear_electric_powers
    p += motor_p4_front_electric_powers
    p += motor_p4_rear_electric_powers
    return p


@sh.add_function(dsp, outputs=['motors_electric_powers'])
def calculate_motors_electric_powers_v1(
        drive_battery_electric_powers, dcdc_converter_electric_powers_demand,
        drive_battery_loads):
    """
    Calculates drive battery load vector [kW].

    :param drive_battery_electric_powers:
        Drive battery electric power [kW].
    :type drive_battery_electric_powers: numpy.array

    :param drive_battery_loads:
        Drive battery load vector [kW].
    :type drive_battery_loads: numpy.array

    :param dcdc_converter_electric_powers_demand:
        DC/DC converter electric power demand [kW].
    :type dcdc_converter_electric_powers_demand: numpy.array

    :return:
        Cumulative motors electric power [kW].
    :rtype: numpy.array
    """
    p = drive_battery_loads - drive_battery_electric_powers
    p += dcdc_converter_electric_powers_demand
    return p



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to model the engine coolant temperature.
"""
import itertools
import numpy as np
import xgboost as xgb
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl
# noinspection SpellCheckingInspection
from sklearn.feature_selection import SelectFromModel
from sklearn.linear_model import RANSACRegressor


def _build_samples(temperature_derivatives, engine_coolant_temperatures, *args):
    col = itertools.chain(
        (engine_coolant_temperatures[:-1],),
        (a[1:] for a in args),
    )
    return np.column_stack(tuple(col)), temperature_derivatives[1:]


# noinspection PyPep8Naming
def _filter_temperature_samples(X, Y, on_engine, thermostat):
    adt = np.abs(Y)
    b = ~((adt <= 0.001) & on_engine[1:])
    b[:co2_utl.argmax(on_engine)] = False
    i = co2_utl.argmax(thermostat < X[:, 0])
    b[i:] = True
    # noinspection PyProtectedMember
    b[:i] &= adt[:i] < dfl.functions._filter_temperature_samples.max_abs_dt_cold
    return X[b], Y[b]


# noinspection PyMissingOrEmptyDocstring
class _SelectFromModel(SelectFromModel):
    def __init__(self, *args, in_mask=(), out_mask=(), **kwargs):
        super(_SelectFromModel, self).__init__(*args, **kwargs)
        self._in_mask = in_mask
        self._out_mask = out_mask
        self._cache_support_mask = None

    def fit(self, *args, **kwargs):
        self._cache_support_mask = None
        return super(_SelectFromModel, self).fit(*args, **kwargs)

    def _get_support_mask(self):
        if self._cache_support_mask is not None:
            return self._cache_support_mask
        if self.prefit:
            estimator = self.estimator
        elif hasattr(self, 'estimator_'):
            estimator = self.estimator_
        else:
            raise ValueError(
                'Either fit the model before transform or set "prefit=True"'
                ' while passing the fitted estimator to the constructor.')
        try:
            with np.errstate(divide='ignore', invalid='ignore'):
                importances = getattr(estimator, "feature_importances_", None)
            if importances is not None and np.isnan(importances).all():
                mask = np.ones(importances.shape, bool)
            else:
                mask = super(_SelectFromModel, self)._get_support_mask()
        except ValueError:
            sfm = SelectFromModel(
                estimator.estimator_, self.threshold, True
            )
            mask = sfm._get_support_mask()

        for i in self._out_mask:
            mask[i] = False

        for i in self._in_mask:
            mask[i] = True
        self._cache_support_mask = mask
        return mask

    def transform(self, X):
        return super(_SelectFromModel, self).transform(X).copy()


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class NoDelta:
    # noinspection PyUnusedLocal
    @staticmethod
    def predict(X, *args):
        return np.zeros(X.shape[0])


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class _SafeRANSACRegressor(RANSACRegressor):
    def fit(self, X, y, **kwargs):
        try:
            return super(_SafeRANSACRegressor, self).fit(X, y, **kwargs)
        except ValueError as ex:
            if self.residual_threshold is None:
                aym = np.abs(y - np.median(y))
                rt = np.median(aym)
                if np.isclose(rt, 0.0):
                    b = ~np.isclose(aym, 0.0)
                    # noinspection PyUnresolvedReferences
                    if b.any():
                        rt = np.median(aym[b])
                self.residual_threshold = rt + np.finfo(np.float32).eps * 10
                res = super(_SafeRANSACRegressor, self).fit(X, y, **kwargs)
                self.residual_threshold = None
                return res
            else:
                raise ex


# noinspection PyMissingOrEmptyDocstring
class _XGBRegressor(xgb.XGBRegressor):
    def __setattr__(self, key, value):
        if key != 'random_state':
            super(_XGBRegressor, self).__setattr__(key, value)
        else:
            super(_XGBRegressor, self).__setattr__(key, 0)

    def cache_params(self):
        self._cache_params = {}

    def get_params(self, deep=True):
        params = getattr(self, '_cache_params', {})
        if deep not in params:
            params[deep] = super(_XGBRegressor, self).get_params(deep)
        return params[deep]


# noinspection PyMethodMayBeStatic,PyMethodMayBeStatic,PyMissingOrEmptyDocstring
class ThermalModel:
    def __init__(self, engine_thermostat_temperature=100.0):
        self.on = self.off = lambda *args: 0
        self.ntemp = 5
        self.thermostat = engine_thermostat_temperature

    # noinspection PyProtectedMember,PyPep8Naming
    def fit(self, engine_coolant_temperatures, engine_temperature_derivatives,
            on_engine, velocities, engine_speeds_out, accelerations,
            after_treatment_warm_up_phases):
        from sklearn.pipeline import Pipeline
        # noinspection PyArgumentEqualDefault
        opt = dict(
            base_estimator=_XGBRegressor(
                random_state=0, objective='reg:squarederror'
            ),
            random_state=0, min_samples=0.85, max_trials=10
        )
        t = np.append(
            [engine_coolant_temperatures[0]], engine_coolant_temperatures[:-1]
        )
        x = np.column_stack((
            velocities, t, np.zeros_like(t), engine_speeds_out, accelerations,
            after_treatment_warm_up_phases
        ))
        n = self.ntemp
        x[np.searchsorted(t, (self.thermostat,))[0]:, 2] = 1
        x[:, 1] = np.round((self.thermostat + n - x[:, 1]) / n) * n
        b = on_engine & (np.abs(engine_temperature_derivatives) > dfl.EPS)
        # noinspection PyArgumentEqualDefault
        model = Pipeline([
            ('selection', _SelectFromModel(
                opt['base_estimator'], threshold='0.8*median', in_mask=(0, 2)
            )),
            ('regression', _SafeRANSACRegressor(**opt))
        ]).fit(x[b, 1:], engine_temperature_derivatives[b])
        model.steps[0][1].estimator_.cache_params()
        model.steps[0][1].estimator.cache_params()
        model.steps[1][1].base_estimator.cache_params()
        model.steps[1][1].estimator_.cache_params()
        self.on = model.predict
        b = ~on_engine
        if b.any():
            model = _SafeRANSACRegressor(**opt).fit(
                x[b, :2].copy(), engine_temperature_derivatives[b]
            )
            model.estimator_.cache_params()
            model.base_estimator.cache_params()
            self.off = model.predict
        return self

    def __call__(self, times, on_engine, velocities, engine_speeds_out,
                 accelerations, after_treatment_warm_up_phases,
                 initial_temperature=23, max_temp=100.0):
        t, temp = initial_temperature, np.zeros_like(times, dtype=float)
        it = enumerate(zip(
            np.ediff1d(times, to_begin=0), on_engine, velocities, accelerations,
            after_treatment_warm_up_phases, engine_speeds_out,
        ))
        x, t0, hot = np.array([[.0] * 6]), self.thermostat + self.ntemp, False
        for i, (dt, b, v, a, atp, s) in it:
            hot |= t > self.thermostat
            x[:] = v, t0 - t, hot, s, a, atp
            t += (self.on(x[:, 1:]) if b else self.off(x[:, :2])) * dt
            temp[i] = t = min(t, max_temp)
        return temp



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to predict the fuel consumptions.
"""
import copy
import functools
import itertools
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl
from numbers import Number

dsp = sh.BlueDispatcher(
    name='Engine fuel consumption sub model',
    description='Calculates fuel consumptions.'
)


@sh.add_function(dsp, outputs=['has_exhausted_gas_recirculation'])
def default_engine_has_exhausted_gas_recirculation(fuel_type):
    """
    Returns the default engine has exhaust gas recirculation value [-].

    :param fuel_type:
        Fuel type (diesel, gasoline, LPG, NG, ethanol, methanol, biodiesel,
        propane).
    :type fuel_type: str

    :return:
        Does the engine have exhaust gas recirculation technology?
    :rtype: bool
    """

    return fuel_type in ('diesel', 'biodiesel')


@sh.add_function(dsp, outputs=['brake_mean_effective_pressures'])
def calculate_brake_mean_effective_pressures(
        engine_speeds_out, engine_powers_out, engine_capacity,
        min_engine_on_speed):
    """
    Calculates engine brake mean effective pressure [bar].

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :return:
        Engine brake mean effective pressure vector [bar].
    :rtype: numpy.array
    """

    b = engine_speeds_out > min_engine_on_speed

    p = np.zeros_like(engine_powers_out)
    p[b] = engine_powers_out[b] / engine_speeds_out[b]
    p[b] *= 1200000.0 / engine_capacity

    return np.nan_to_num(p)


@sh.add_function(dsp, outputs=['extended_integration_times'])
def calculate_extended_integration_times(
        times, velocities, on_engine, phases_integration_times, stop_velocity,
        after_treatment_warm_up_phases):
    """
    Calculates the extended integration times [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param phases_integration_times:
        Cycle phases integration times [s].
    :type phases_integration_times: tuple

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Extended cycle phases integration times [s].
    :rtype: tuple
    """
    from ..gear_box.mechanical import _shift
    lv, pit = np.zeros(velocities.size + 2), np.unique(phases_integration_times)
    lv[1:-1] = np.asarray(velocities <= stop_velocity, int)
    indices = np.where(np.diff(lv) != 0)[0].reshape(-1, 2)
    split_points = []
    for i, j in indices:
        t0, t1 = times[i], times[j - 1]
        if t1 - t0 < 20 or any(t0 <= x <= t1 for x in pit):
            continue

        b = ~on_engine[i:j]
        if b.any() and not b.all():
            t = np.median(times[i:j][b])
        else:
            t = (t0 + t1) / 2
        split_points.append(t)
    try:
        for i in _shift(after_treatment_warm_up_phases):
            if not lv[i + 1]:
                split_points.append(times[i])
    except IndexError:
        pass

    return sorted(split_points)


@sh.add_function(dsp, outputs=[
    'extended_cumulative_co2_emissions', 'extended_phases_integration_times'
])
def calculate_extended_cumulative_co2_emissions(
        times, velocities, on_engine, extended_integration_times,
        co2_normalization_references, phases_integration_times,
        phases_co2_emissions, phases_distances, stop_velocity):
    """
    Calculates the extended cumulative CO2 of cycle phases [CO2g].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param extended_integration_times:
        Extended cycle phases integration times [s].
    :type extended_integration_times: tuple

    :param co2_normalization_references:
        CO2 normalization references (e.g., engine loads) [-].
    :type co2_normalization_references: numpy.array

    :param phases_integration_times:
        Cycle phases integration times [s].
    :type phases_integration_times: tuple

    :param phases_co2_emissions:
        CO2 emission of cycle phases [CO2g/km].
    :type phases_co2_emissions: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Extended cumulative CO2 of cycle phases [CO2g].
    :rtype: numpy.array
    """

    r = co2_normalization_references.copy()
    r[~on_engine] = 0
    lv = np.asarray(velocities <= stop_velocity, int)
    _cco2, phases = [], []
    cco2 = phases_co2_emissions * phases_distances
    from scipy.integrate import trapz

    def _stops(n, m):
        return trapz(lv[n:m], times[n:m]) / (times[m] - times[n])

    for cco2, (t0, t1) in zip(cco2, phases_integration_times):
        i, j = np.searchsorted(times, (t0, t1))
        if i == j:
            continue
        v = trapz(r[i:j], times[i:j])
        c, k0 = [0.0], i

        p = [t for t in extended_integration_times if t0 < t < t1]
        for k, t in zip(np.searchsorted(times, p), p):
            if k < j and _stops(k0, k) < 0.5 and _stops(k, j) < 0.5:
                phases.append((t0, t))
                t0, k0 = t, k
                c.append(trapz(r[i:k], times[i:k]) / v)
        phases.append((t0, t1))
        c.append(1.0)

        _cco2.extend(np.diff(c) * cco2)

    return np.array(_cco2), phases


# noinspection PyMissingOrEmptyDocstring
class IdleFuelConsumptionModel:
    def __init__(self, fc=None):
        self.fc = fc
        self.n_s = None
        self.c = None
        self.fmep_model = None

    def fit(self, idle_engine_speed, engine_capacity, engine_stroke, lhv,
            fmep_model):
        idle = idle_engine_speed[0]
        from . import calculate_mean_piston_speeds
        self.n_s = calculate_mean_piston_speeds(idle, engine_stroke)
        self.c = idle * (engine_capacity / (lhv * 1200))
        self.fmep_model = fmep_model
        return self

    def consumption(self, params=None, ac_phases=None):
        import lmfit
        if isinstance(params, lmfit.Parameters):
            params = params.valuesdict()

        if self.fc is not None:
            ac = params.get('acr', self.fmep_model.base_acr)
            avv = params.get('avv', 0)
            lb = params.get('lb', 0)
            egr = params.get('egr', 0)
            fc = self.fc
        else:
            fc, _, ac, avv, lb, egr = self.fmep_model(params, self.n_s, 0, 1)

            if not (ac_phases is None or ac_phases.all() or 'acr' in params):
                p = params.copy()
                p['acr'] = self.fmep_model.base_acr
                _fc, _, _ac, _avv, _lb, _egr = self.fmep_model(p, self.n_s, 0,
                                                               1)
                fc = np.where(ac_phases, fc, _fc)
                ac = np.where(ac_phases, ac, _ac)
                avv = np.where(ac_phases, avv, _avv)
                lb = np.where(ac_phases, lb, _lb)
                egr = np.where(ac_phases, egr, _egr)

            fc *= self.c  # [g/sec]
        return fc, ac, avv, lb, egr


dsp.add_data('stop_velocity', dfl.values.stop_velocity)
dsp.add_data('min_engine_on_speed', dfl.values.min_engine_on_speed)


@sh.add_function(
    dsp, inputs_kwargs=True, inputs_defaults=True,
    outputs=['idle_fuel_consumption_model']
)
def define_idle_fuel_consumption_model(
        idle_engine_speed, engine_capacity, engine_stroke,
        engine_fuel_lower_heating_value, fmep_model,
        idle_fuel_consumption_initial_guess=None):
    """
    Defines the idle fuel consumption model.

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :param idle_fuel_consumption_initial_guess:
        Initial guess of fuel consumption at hot idle engine speed [g/s].
    :type idle_fuel_consumption_initial_guess: float, optional

    :return:
        Idle fuel consumption model.
    :rtype: IdleFuelConsumptionModel
    """
    return IdleFuelConsumptionModel(idle_fuel_consumption_initial_guess).fit(
        idle_engine_speed, engine_capacity, engine_stroke,
        engine_fuel_lower_heating_value, fmep_model
    )


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['engine_idle_fuel_consumption']
)
def calculate_engine_idle_fuel_consumption(
        idle_fuel_consumption_model, co2_params_calibrated=None):
    """
    Calculates fuel consumption at hot idle engine speed [g/s].

    :param idle_fuel_consumption_model:
        Idle fuel consumption model.
    :type idle_fuel_consumption_model: IdleFuelConsumptionModel

    :param co2_params_calibrated:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type co2_params_calibrated: lmfit.Parameters

    :return:
        Fuel consumption at hot idle engine speed [g/s].
    :rtype: float
    """

    return idle_fuel_consumption_model.consumption(co2_params_calibrated)[0]


def _yield_factors(param_id, factor):
    try:
        for k, v in factor.get(param_id, {}).items():
            yield k, v, 1
    except TypeError:
        p = {}

        def _defaults():
            return (np.zeros_like(param_id, dtype=float),
                    np.zeros_like(param_id, dtype=int))

        for m in np.unique(param_id):
            if not isinstance(m, np.ma.core.MaskedConstant):
                b = m == param_id
                for k, v in factor.get(m, {}).items():
                    j, i = sh.get_nested_dicts(p, k, default=_defaults)
                    j[b], i[b] = v, 1

        for k, (j, n) in p.items():
            b = n == 0
            j[b], n[b] = 1, 1
            yield k, j, n


def _tech_mult_factors(**params):
    p = {}
    # noinspection PyProtectedMember
    factors = dfl.functions._tech_mult_factors.factors
    for k, v in factors.items():
        for i, j, n in _yield_factors(params.get(k, 0), v):
            s = sh.get_nested_dicts(p, i, default=lambda: [0, 0])
            s[0] += j
            s[1] += n

    for k, (n, d) in p.items():
        m = n / d
        params[k] = m * params[k]

    return params


# noinspection PyUnusedLocal,PyPep8Naming
def _ABC(
        n_speeds, n_powers=0, n_temperatures=1,
        a2=0, b2=0, a=0, b=0, c=0, t=0, l=0, l2=0, acr=1, **kw):
    acr2 = (acr ** 2)
    A = a2 / acr2 + (b2 / acr2) * n_speeds
    B = a / acr + (b / acr + (c / acr) * n_speeds) * n_speeds
    C = l + l2 * n_speeds ** 2
    if not isinstance(n_temperatures, Number) or n_temperatures != 1:
        C *= np.power(n_temperatures, -t)
    C -= n_powers / acr

    return A, B, C


# noinspection PyPep8Naming
def _fuel_ABC(n_speeds, **kw):
    with np.errstate(under='ignore'):
        return _ABC(n_speeds, **_tech_mult_factors(**kw))


# noinspection PyPep8Naming
def _calculate_fc(A, B, C):
    with np.errstate(under='ignore'):
        b = np.array(A, dtype=bool)
        if b.all():
            v = np.sqrt(np.abs(B ** 2 - 4.0 * A * C))
            return (-B + v) / (2 * A), v
        elif ~b.all():
            return -C / B, B
        else:
            fc, v = np.zeros_like(C), np.zeros_like(C)
            fc[b], v[b] = _calculate_fc(A[b], B[b], C[b])
            b = ~b
            fc[b], v[b] = _calculate_fc(A[b], B[b], C[b])
            return fc, v


# noinspection PyMissingOrEmptyDocstring
class FMEP:
    def __init__(self, full_bmep_curve, active_cylinder_ratios=(1.0,),
                 has_cylinder_deactivation=False,
                 acr_full_bmep_curve_percentage=0.5,
                 acr_max_mean_piston_speeds=12.0,
                 acr_min_mean_piston_speeds=3.0,
                 has_variable_valve_actuation=False,
                 has_lean_burn=False,
                 lb_max_mean_piston_speeds=12.0,
                 lb_full_bmep_curve_percentage=0.4,
                 has_exhausted_gas_recirculation=False,
                 has_selective_catalytic_reduction=False,
                 egr_max_mean_piston_speeds=12.0,
                 egr_full_bmep_curve_percentage=0.7,
                 engine_type=None):

        if active_cylinder_ratios:
            self.base_acr = max(active_cylinder_ratios)
        else:
            self.base_acr = 1.0
        self.active_cylinder_ratios = set(active_cylinder_ratios)
        self.active_cylinder_ratios -= {self.base_acr}
        self.fbc = full_bmep_curve

        self.has_cylinder_deactivation = has_cylinder_deactivation
        self.acr_max_mean_piston_speeds = float(acr_max_mean_piston_speeds)
        self.acr_min_mean_piston_speeds = float(acr_min_mean_piston_speeds)
        self.acr_fbc_percentage = acr_full_bmep_curve_percentage

        self.has_variable_valve_actuation = has_variable_valve_actuation

        self.has_lean_burn = has_lean_burn
        self.lb_max_mean_piston_speeds = float(lb_max_mean_piston_speeds)
        self.lb_fbc_percentage = lb_full_bmep_curve_percentage
        self.lb_n_temp_min = 0.5

        self.has_exhausted_gas_recirculation = has_exhausted_gas_recirculation
        self.has_selective_catalytic_reduction = \
            has_selective_catalytic_reduction
        self.egr_max_mean_piston_speeds = float(egr_max_mean_piston_speeds)
        self.egr_fbc_percentage = egr_full_bmep_curve_percentage
        self.engine_type = engine_type

    def vva(self, params, n_powers, a=None):
        a = a or {}
        if self.has_variable_valve_actuation and 'vva' not in params:
            a['vva'] = ((0, True), (1, n_powers >= 0))
        return a

    def lb(self, params, n_speeds, n_powers, n_temp, a=None):
        a = a or {}
        if self.has_lean_burn and 'lb' not in params:
            b = n_temp >= self.lb_n_temp_min
            b &= n_speeds < self.lb_max_mean_piston_speeds
            b &= n_powers <= (self.fbc(n_speeds) * self.lb_fbc_percentage)
            a['lb'] = ((0, True), (1, b))
        return a

    def egr(self, params, n_speeds, n_powers, n_temp, a=None):
        a = a or {}
        if self.has_exhausted_gas_recirculation and 'egr' not in params:
            # b = n_speeds < self.egr_max_mean_piston_speeds
            # b &= n_powers <= (self.fbc(n_speeds) * self.egr_fbc_percentage)
            k = self.engine_type, self.has_selective_catalytic_reduction
            egr = dfl.functions.FMEP_egr.egr_fact_map[k]
            if k[0] == 'compression':
                if k[1]:
                    b = n_temp < 1
                else:
                    b = n_speeds < self.egr_max_mean_piston_speeds
                    b &= n_powers <= (
                            self.fbc(n_speeds) * self.egr_fbc_percentage
                    )

                if b is True:
                    a['egr'] = (egr, True),
                elif b is False:
                    a['egr'] = (0, True),
                else:
                    a['egr'] = (np.where(b, egr, 0), True),
            else:
                a['egr'] = ((0, True), (egr, True))

        return a

    def acr(self, params, n_speeds, n_powers, n_temp, a=None, acr_valid=None):
        a = a or {'acr': [(self.base_acr, True)]}
        if self.has_cylinder_deactivation and self.active_cylinder_ratios and \
                'acr' not in params:
            l = a['acr']
            b = n_powers > 0
            b &= (self.acr_min_mean_piston_speeds < n_speeds)
            b &= (n_speeds < self.acr_max_mean_piston_speeds)
            if acr_valid is not None:
                b &= acr_valid
            ac = self.fbc(n_speeds) * self.acr_fbc_percentage
            ac = n_powers / np.maximum(dfl.EPS, ac)
            for acr in sorted(self.active_cylinder_ratios):
                l.append((acr, b & (ac < acr)))
        return a

    @staticmethod
    def _check_combinations(a):
        out = {}
        for k, v in a.items():
            for i in v:
                try:
                    if i[1] is True or i[1].any():
                        sh.get_nested_dicts(out, k, default=list).append(i)
                except AttributeError:
                    pass
        return out

    def combination(self, params, n_speeds, n_powers, n_temp, acr_valid=None):
        a = self.acr(params, n_speeds, n_powers, n_temp, acr_valid=acr_valid)
        a = self.lb(params, n_speeds, n_powers, n_temp, a=a)
        a = self.vva(params, n_powers, a=a)
        a = self.egr(params, n_speeds, n_powers, n_temp, a=a)
        a = self._check_combinations(a)

        keys, c = zip(*sorted(a.items()))
        p = params.copy()
        p.update({
            'n_speeds': n_speeds,
            'n_powers': n_powers,
            'n_temperatures': n_temp
        })
        for s in itertools.product(*c):
            b, d = True, {}

            for k, (v, n) in zip(keys, s):
                b &= n
                d[k] = v
            try:
                if b is False or not b.any():
                    continue
                if b.all():
                    b = True
            except AttributeError:
                pass

            p.update(d)
            yield {k: self.g(v, b) for k, v in p.items()}, d, b

    @staticmethod
    def g(data, b):
        if b is not True:
            try:
                return np.ma.masked_where(~b, data, copy=False)
            except (IndexError, TypeError):
                pass
        return data

    def __call__(self, params, n_speeds, n_powers, n_temp, acr_valid=None):
        it = self.combination(params, n_speeds, n_powers, n_temp, acr_valid)
        s = None
        for p, d, n in it:
            d['fmep'], d['v'] = _calculate_fc(*_fuel_ABC(**p))
            if s is None:
                s = d
            else:
                b = d['fmep'] < s['fmep']
                if n is True and b is True:
                    s = d
                elif b is not False:
                    n &= b
                    if n.all():
                        s = d
                    elif n.any():
                        for k, v in d.items():
                            s[k] = np.where(n, v, s[k])

        acr = s.get('acr', params.get('acr', self.base_acr))
        vva = s.get('vva', params.get('vva', 0))
        lb = s.get('lb', params.get('lb', 0))
        egr = s.get('egr', params.get('egr', 0))

        return s['fmep'], s['v'], acr, vva, lb, egr


@sh.add_function(dsp, outputs=['cylinder_deactivation_valid_phases'])
def calculate_cylinder_deactivation_valid_phases(engine_inertia_powers_losses):
    """
    Calculates valid activation phases for cylinder deactivation.

    :param engine_inertia_powers_losses:
        Engine power losses due to inertia [kW].
    :type engine_inertia_powers_losses: numpy.array

    :return:
        Valid activation phases for cylinder deactivation.
    :rtype: numpy.array
    """
    p = dfl.functions.calculate_cylinder_deactivation_valid_phases.LIMIT
    return engine_inertia_powers_losses <= p


dsp.add_data('active_cylinder_ratios', dfl.values.active_cylinder_ratios)
dsp.add_data(
    'engine_has_cylinder_deactivation',
    dfl.values.engine_has_cylinder_deactivation
)
dsp.add_data(
    'engine_has_variable_valve_actuation',
    dfl.values.engine_has_variable_valve_actuation
)
dsp.add_data('has_lean_burn', dfl.values.has_lean_burn)
dsp.add_data(
    'has_selective_catalytic_reduction',
    dfl.values.has_selective_catalytic_reduction
)


@sh.add_function(dsp, outputs=['fmep_model'])
def define_fmep_model(
        full_bmep_curve, engine_max_speed, engine_stroke,
        active_cylinder_ratios, engine_has_cylinder_deactivation,
        engine_has_variable_valve_actuation, has_lean_burn,
        has_exhausted_gas_recirculation, has_selective_catalytic_reduction,
        engine_type, idle_engine_speed):
    """
    Defines the vehicle FMEP model.

    :param full_bmep_curve:
        Vehicle full bmep curve.
    :type full_bmep_curve: function

    :param engine_max_speed:
        Maximum allowed engine speed [RPM].
    :type engine_max_speed: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param active_cylinder_ratios:
        Possible active cylinder ratios [-].
    :type active_cylinder_ratios: tuple[float]

    :param engine_has_cylinder_deactivation:
        Does the engine have cylinder deactivation technology?
    :type engine_has_cylinder_deactivation: bool

    :param engine_has_variable_valve_actuation:
        Does the engine feature variable valve actuation? [-].
    :type engine_has_variable_valve_actuation: bool

    :param has_lean_burn:
        Does the engine have lean burn technology?
    :type has_lean_burn: bool

    :param has_exhausted_gas_recirculation:
        Does the engine have exhaust gas recirculation technology?
    :type has_exhausted_gas_recirculation: bool

    :param has_selective_catalytic_reduction:
        Does the engine have selective catalytic reduction technology?
    :type has_selective_catalytic_reduction: bool

    :param engine_type:
        Engine type (positive turbo, positive natural aspiration, compression).
    :type engine_type: str

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Vehicle FMEP model.
    :rtype: FMEP
    """

    d = dfl.functions.define_fmep_model
    acr_fbcp = d.acr_full_bmep_curve_percentage
    lb_fbcp = d.lb_full_bmep_curve_percentage
    egr_fbcp = d.egr_full_bmep_curve_percentage

    acr_maps = d.acr_max_mean_piston_speeds_percentage * engine_max_speed
    acr_mips = d.acr_min_mean_piston_speeds_percentage * idle_engine_speed[0]

    lb_mps = d.lb_max_mean_piston_speeds_percentage * engine_max_speed
    egr_mps = d.egr_max_mean_piston_speeds_percentage * engine_max_speed

    from . import calculate_mean_piston_speeds
    bmep = calculate_mean_piston_speeds

    model = FMEP(
        full_bmep_curve,
        active_cylinder_ratios=active_cylinder_ratios,
        has_cylinder_deactivation=engine_has_cylinder_deactivation,
        acr_full_bmep_curve_percentage=acr_fbcp,
        acr_max_mean_piston_speeds=bmep(acr_maps, engine_stroke),
        acr_min_mean_piston_speeds=bmep(acr_mips, engine_stroke),
        has_variable_valve_actuation=engine_has_variable_valve_actuation,
        has_lean_burn=has_lean_burn,
        lb_max_mean_piston_speeds=bmep(lb_mps, engine_stroke),
        lb_full_bmep_curve_percentage=lb_fbcp,
        has_exhausted_gas_recirculation=has_exhausted_gas_recirculation,
        has_selective_catalytic_reduction=has_selective_catalytic_reduction,
        egr_max_mean_piston_speeds=bmep(egr_mps, engine_stroke),
        egr_full_bmep_curve_percentage=egr_fbcp,
        engine_type=engine_type
    )

    return model


# noinspection PyUnresolvedReferences
@sh.add_function(dsp, outputs=['fuel_map'])
def define_fuel_map(
        idle_engine_speed, engine_capacity, co2_params_calibrated, fmep_model,
        engine_fuel_lower_heating_value, engine_stroke, full_load_speeds,
        full_load_powers):
    """
    Define fuel consumption map [RPM, kW, g/s].

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :param full_load_speeds:
        T1 map speed vector [RPM].
    :type full_load_speeds: numpy.array

    :param full_load_powers:
        T1 map power vector [kW].
    :type full_load_powers: numpy.array

    :param co2_params_calibrated:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type co2_params_calibrated: lmfit.Parameters

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :return:
        Fuel consumption map [RPM, kW, g/s].
    :rtype: dict
    """
    from . import calculate_mean_piston_speeds
    speed = np.linspace(full_load_speeds[0], full_load_speeds[-1], 100).tolist()
    speed = np.unique(speed + list(full_load_speeds))
    p = co2_params_calibrated.valuesdict()
    lhv = engine_fuel_lower_heating_value
    par = dfl.functions.calculate_co2_emissions
    idle_cutoff = idle_engine_speed[0] * par.cutoff_idle_ratio
    ec_p0 = _calculate_p0(
        fmep_model, p, engine_capacity, engine_stroke, idle_cutoff, lhv
    )
    flp = np.interp(speed, full_load_speeds, full_load_powers)
    power = np.linspace(ec_p0, np.max(full_load_powers), 100).tolist()
    power = np.unique(power + list(full_load_powers) + list(flp) + [0])

    e_s, e_p = np.meshgrid(speed, power, indexing='ij')
    n_s = calculate_mean_piston_speeds(e_s, engine_stroke)
    n_p = calculate_brake_mean_effective_pressures(e_s, e_p, engine_capacity, 0)

    fc = np.maximum(0, fmep_model(p, n_s, n_p, 1)[0])
    fc *= e_s * (engine_capacity / (lhv * 1200))  # [g/sec]
    fc[np.argmax(fc, axis=1)[:, None] < np.arange(fc.shape[1])] = np.nan
    b = ~np.isnan(fc).all(0)
    return dict(
        speed=speed.tolist(), power=power[b].tolist(), fuel=fc[:, b].tolist()
    )


def _calculate_p0(
        fmep_model, params, engine_capacity, engine_stroke,
        idle_engine_speed_median, engine_fuel_lower_heating_value):
    """
    Calculates the engine power threshold limit [kW].

    :param params:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type params: dict

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param idle_engine_speed_median:
        Engine speed idle median [RPM].
    :type idle_engine_speed_median: float

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :return:
        Engine power threshold limit [kW].
    :rtype: float
    """

    engine_cm_idle = idle_engine_speed_median * engine_stroke / 30000.0

    lhv = engine_fuel_lower_heating_value

    wfb_idle, wfa_idle = fmep_model(params, engine_cm_idle, 0, 1)[:2]
    wfa_idle = (3600000.0 / lhv) / wfa_idle
    wfb_idle *= (3.0 * engine_capacity / lhv * idle_engine_speed_median)
    return -wfb_idle / wfa_idle


def _apply_ac_phases(func, fmep_model, params, *args, ac_phases=None):
    res_with_ac = func(fmep_model, params, *args)
    if ac_phases is None or ac_phases.all() or not (
            fmep_model.has_cylinder_deactivation and
            fmep_model.active_cylinder_ratios):
        return res_with_ac
    else:
        p = params.copy()
        p['acr'] = fmep_model.base_acr
        return np.where(ac_phases, res_with_ac, func(fmep_model, p, *args))


def _normalized_engine_temperatures(
        engine_temperatures, temperature_target):
    """
    Calculates the normalized engine coolant temperatures [-].

    ..note::
        Engine coolant temperatures are first converted in kelvin and then
        normalized. The results is between ``[0, 1]``.

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param temperature_target:
        Normalization temperature [°C].
    :type temperature_target: float

    :return:
        Normalized engine coolant temperature [-].
    :rtype: numpy.array
    """
    temp = (engine_temperatures + 273.0) / (temperature_target + 273.0)
    if isinstance(temp, np.ndarray):
        temp[np.searchsorted(temp, (1,))[0]:] = 1
    return np.minimum(1, temp)


def _calculate_co2_emissions(
        time_series, engine_fuel_lower_heating_value, idle_engine_speed,
        engine_stroke, engine_capacity, idle_fuel_consumption_model,
        fuel_carbon_content, min_engine_on_speed, fmep_model, params,
        sub_values=None):
    """
    Calculates CO2 emissions [CO2g/s].

    :param time_series:
        Engine speed vector [RPM], Engine power vector [kW], Engine coolant
        temperature vector [°C], Mean piston speed vector [m/s], and Engine
        brake mean effective pressure vector [bar].
    :type time_series: numpy.array

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param idle_fuel_consumption_model:
        Model of fuel consumption at hot idle engine speed.
    :type idle_fuel_consumption_model: IdleFuelConsumptionModel

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :param params:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type params: lmfit.Parameters

    :param sub_values:
        Boolean vector.
    :type sub_values: numpy.array, optional

    :return:
        CO2 emissions vector [CO2g/s].
    :rtype: numpy.array
    """

    p = params.valuesdict()
    # namespace shortcuts
    if sub_values is not None:
        e_s, e_p, e_t, n_s, n_p, acr_v, wu_p = time_series[:, sub_values]
    else:
        e_s, e_p, e_t, n_s, n_p, acr_v, wu_p = time_series
    lhv, acr_v = engine_fuel_lower_heating_value, acr_v.astype(bool)
    wu_p = wu_p.astype(bool)
    idle_fc_model = idle_fuel_consumption_model.consumption
    fc, ac, vva, lb, egr = np.zeros((5, len(e_p)), dtype=float)
    ac[:] = 1
    # Idle fc correction for temperature
    n = (e_s < idle_engine_speed[0] + min_engine_on_speed)
    _b = (e_s >= min_engine_on_speed)
    par = dfl.functions.calculate_co2_emissions
    idle_cutoff = idle_engine_speed[0] * par.cutoff_idle_ratio

    if p['t0'] == 0 and p['t1'] == 0:
        ac_phases, n_t = np.ones_like(e_p, dtype=bool), 1
        ec_p0 = _calculate_p0(
            fmep_model, p, engine_capacity, engine_stroke, idle_cutoff, lhv
        )
        _b &= ~((e_p <= ec_p0) & (e_s > idle_cutoff))
        b = n & _b
        fc[b], ac[b], vva[b], lb[b], egr[b] = idle_fc_model(p)
        b = ~n & _b
    else:
        p['t'] = np.where(wu_p, p['t0'], p['t1'])
        n_t = _normalized_engine_temperatures(e_t, p['trg'])
        ac_phases = ~wu_p & acr_v
        ec_p0 = _apply_ac_phases(
            _calculate_p0, fmep_model, p, engine_capacity, engine_stroke,
            idle_cutoff, lhv, ac_phases=ac_phases
        )
        _b &= ~((e_p <= ec_p0) & (e_s > idle_cutoff))
        b = n & _b
        # noinspection PyUnresolvedReferences
        idle_fc, ac[b], vva[b], lb[b], egr[b] = idle_fc_model(p, ac_phases[b])
        fc[b] = idle_fc * np.power(n_t[b], -p['t'][b])
        b = ~n & _b
        p['t'] = p['t'][b]
        n_t = n_t[b]

    fc[b], _, ac[b], vva[b], lb[b], egr[b] = fmep_model(
        p, n_s[b], n_p[b], n_t, acr_valid=acr_v[b]
    )
    fc[b] *= e_s[b] * (engine_capacity / (lhv * 1200))  # [g/sec]
    fc[fc < 0] = 0

    co2 = fc * fuel_carbon_content

    return np.nan_to_num(co2), ac, vva, lb, egr


# noinspection PyUnusedLocal
@sh.add_function(dsp, outputs=['co2_emissions_model'])
def define_co2_emissions_model(
        engine_speeds_out, engine_powers_out, mean_piston_speeds,
        brake_mean_effective_pressures, engine_temperatures, on_engine,
        cylinder_deactivation_valid_phases, after_treatment_warm_up_phases,
        engine_fuel_lower_heating_value, idle_engine_speed, engine_stroke,
        engine_capacity, idle_fuel_consumption_model, fuel_carbon_content,
        min_engine_on_speed, fmep_model):
    """
    Returns CO2 emissions model (see :func:`calculate_co2_emissions`).

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param mean_piston_speeds:
        Mean piston speed vector [m/s].
    :type mean_piston_speeds: numpy.array

    :param brake_mean_effective_pressures:
        Engine brake mean effective pressure vector [bar].
    :type brake_mean_effective_pressures: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param cylinder_deactivation_valid_phases:
        Valid activation phases for cylinder deactivation.
    :type cylinder_deactivation_valid_phases: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine speed is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param idle_fuel_consumption_model:
        Idle fuel consumption model.
    :type idle_fuel_consumption_model: IdleFuelConsumptionModel

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :return:
        CO2 emissions model (co2_emissions = models(params)).
    :rtype: callable
    """

    ts = (
        engine_speeds_out, engine_powers_out, engine_temperatures,
        mean_piston_speeds, brake_mean_effective_pressures,
        cylinder_deactivation_valid_phases, after_treatment_warm_up_phases
    )

    model = functools.partial(
        _calculate_co2_emissions, np.array(ts, copy=False),
        engine_fuel_lower_heating_value, idle_engine_speed, engine_stroke,
        engine_capacity, idle_fuel_consumption_model, fuel_carbon_content,
        min_engine_on_speed, fmep_model
    )

    return model


# noinspection PyUnusedLocal
@sh.add_function(
    dsp, outputs=['co2_emissions_model'], input_domain=co2_utl.check_first_arg
)
def define_co2_emissions_model_hybrid_calibration(
        is_hybrid, engine_speeds_out, engine_powers_out, mean_piston_speeds,
        brake_mean_effective_pressures, engine_temperatures, on_engine,
        cylinder_deactivation_valid_phases, times,
        engine_thermostat_temperature, is_cycle_hot,
        engine_fuel_lower_heating_value, idle_engine_speed, engine_stroke,
        engine_capacity, idle_fuel_consumption_model, fuel_carbon_content,
        min_engine_on_speed, fmep_model):
    """
    Returns CO2 emissions model (see :func:`calculate_co2_emissions`).

    :param is_hybrid:
        Is the vehicle hybrid?
    :type is_hybrid: bool

    :param engine_speeds_out:
        Engine speed vector [RPM].
    :type engine_speeds_out: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param mean_piston_speeds:
        Mean piston speed vector [m/s].
    :type mean_piston_speeds: numpy.array

    :param brake_mean_effective_pressures:
        Engine brake mean effective pressure vector [bar].
    :type brake_mean_effective_pressures: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param cylinder_deactivation_valid_phases:
        Valid activation phases for cylinder deactivation.
    :type cylinder_deactivation_valid_phases: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_thermostat_temperature:
        Engine thermostat temperature [°C].
    :type engine_thermostat_temperature: float

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool

    :param engine_fuel_lower_heating_value:
        Fuel lower heating value [kJ/kg].
    :type engine_fuel_lower_heating_value: float

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :param engine_stroke:
        Engine stroke [mm].
    :type engine_stroke: float

    :param engine_capacity:
        Engine capacity [cm3].
    :type engine_capacity: float

    :param idle_fuel_consumption_model:
        Idle fuel consumption model.
    :type idle_fuel_consumption_model: IdleFuelConsumptionModel

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :param fmep_model:
        Engine FMEP model.
    :type fmep_model: FMEP

    :return:
        CO2 emissions model (co2_emissions = models(params)).
    :rtype: callable
    """
    # noinspection PyProtectedMember
    from ..control.hybrid import _filter_warm_up
    after_treatment_warm_up_phases = _filter_warm_up(
        times, on_engine, on_engine, engine_thermostat_temperature,
        is_cycle_hot, engine_temperatures
    )
    return define_co2_emissions_model(
        engine_speeds_out, engine_powers_out, mean_piston_speeds,
        brake_mean_effective_pressures, engine_temperatures, on_engine,
        cylinder_deactivation_valid_phases, after_treatment_warm_up_phases,
        engine_fuel_lower_heating_value, idle_engine_speed, engine_stroke,
        engine_capacity, idle_fuel_consumption_model, fuel_carbon_content,
        min_engine_on_speed, fmep_model
    )


def _select_initial_friction_params(co2_params_initial_guess):
    """
    Selects initial guess of friction params l & l2 for the calculation of
    the motoring curve.

    :param co2_params_initial_guess:
        Initial guess of CO2 emission model params.
    :type co2_params_initial_guess: lmfit.Parameters

    :return:
        Initial guess of friction params l & l2.
    :rtype: float, float
    """

    params = co2_params_initial_guess.valuesdict()

    return sh.selector(('l', 'l2'), params, output_type='list')


# noinspection PyUnusedLocal
def _missing_co2_params(params, *args, _not=False):
    """
    Checks if all co2_params are not defined.

    :param params:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).
    :type params: dict | lmfit.Parameters

    :param _not:
        If True the function checks if not all co2_params are defined.
    :type _not: bool

    :return:
        If is missing some parameter.
    :rtype: bool
    """

    s = {'a', 'b', 'c', 'a2', 'b2', 'l', 'l2', 't0', 'dt', 'trg'}

    if _not:
        return set(params).issuperset(s)

    return not set(params).issuperset(s)


@sh.add_function(
    dsp, inputs_kwargs=True, inputs_defaults=True,
    outputs=['co2_params_initial_guess', 'initial_friction_params']
)
def define_initial_co2_emission_model_params_guess(
        co2_params, engine_type, engine_thermostat_temperature,
        engine_thermostat_temperature_window,
        engine_n_cylinders=dfl.values.engine_n_cylinders,
        is_cycle_hot=dfl.values.is_cycle_hot):
    """
    Selects initial guess and bounds of co2 emission model params.

    :param co2_params:
        CO2 emission model params (a2, b2, a, b, c, l, l2, t, trg).
    :type co2_params: dict

    :param engine_type:
        Engine type (positive turbo, positive natural aspiration, compression).
    :type engine_type: str

    :param engine_thermostat_temperature:
        Engine normalization temperature [°C].
    :type engine_thermostat_temperature: float

    :param engine_thermostat_temperature_window:
        Thermostat engine temperature limits [°C].
    :type engine_thermostat_temperature_window: (float, float)

    :param engine_n_cylinders:
        Number of engine cylinders [-].
    :type engine_n_cylinders: int

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool, optional

    :return:
        Initial guess of co2 emission model params and of friction params.
    :rtype: lmfit.Parameters, list[float]
    """
    import lmfit
    import collections
    bounds = {}  # Parameters bounds.
    par = dfl.functions.define_initial_co2_emission_model_params_guess
    default = collections.OrderedDict(
        copy.deepcopy(par.CO2_PARAMS[engine_type])
    )
    default['trg'] = {
        'value': engine_thermostat_temperature,
        'min': engine_thermostat_temperature_window[0],
        'max': engine_thermostat_temperature_window[1],
        'vary': False
    }

    keys, n = ('l', 'l2'), engine_n_cylinders / default.pop('n_cylinders', 4)
    for d in sh.selector(keys, default, allow_miss=True).values():
        for k in {'value', 'min', 'max'}.intersection(d):
            d[k] *= n

    if is_cycle_hot:
        default['t1'].update({'value': 0.0, 'vary': False})
        default['dt'].update({'value': 0.0, 'vary': False})

    p = lmfit.Parameters()
    eps = dfl.EPS
    for k, kw in default.items():
        kw['name'] = k

        kw['value'] = co2_params.get(k, kw.get('value', None))

        if k in bounds:
            b = bounds[k]
            kw['min'] = b.get('min', kw.get('min', None))
            kw['max'] = b.get('max', kw.get('max', None))
            kw['vary'] = b.get('vary', kw.get('vary', True))
        elif 'vary' not in kw:
            kw['vary'] = k not in co2_params

        if kw['value'] is not None:
            if 'min' in kw and kw['value'] < kw['min']:
                kw['min'] = kw['value'] - eps
            if 'max' in kw and kw['value'] > kw['max']:
                kw['max'] = kw['value'] + eps

        if 'min' in kw and 'max' in kw and kw['min'] == kw['max']:
            kw['vary'] = False
            # noinspection PyTypeChecker
            kw['max'] = kw['min'] = None

        kw['min'] = kw.get('min', None)
        kw['max'] = kw.get('max', None)
        p.add(**kw)

    friction_params = _select_initial_friction_params(p)
    if not _missing_co2_params(co2_params):
        p = sh.NONE

    return p, friction_params


@sh.add_function(dsp, outputs=['extended_phases_distances'])
def calculate_extended_phases_distances(
        times, extended_phases_integration_times, velocities):
    """
    Calculates extended cycle phases distances [km].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param extended_phases_integration_times:
        Extended cycle phases integration times [s].
    :type extended_phases_integration_times: tuple

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :return:
        Extended cycle phases distances [km].
    :rtype: numpy.array
    """
    from ..co2 import calculate_phases_distances, identify_phases_indices
    indices = identify_phases_indices(times, extended_phases_integration_times)
    return calculate_phases_distances(indices, velocities)


@sh.add_function(dsp, outputs=['extended_phases_co2_emissions'])
def calculate_extended_phases_co2_emissions(
        extended_cumulative_co2_emissions, extended_phases_distances):
    """
    Calculates the extended CO2 emission of cycle phases [CO2g/km].

    :param extended_cumulative_co2_emissions:
        Extended cumulative CO2 of cycle phases [CO2g].
    :type extended_cumulative_co2_emissions: numpy.array

    :param extended_phases_distances:
        Extended cycle phases distances [km].
    :type extended_phases_distances: numpy.array

    :return:
        Extended CO2 emission of cycle phases [CO2g/km].
    :rtype: numpy.array
    """
    with np.errstate(divide='ignore', invalid='ignore'):
        x = extended_cumulative_co2_emissions
        return np.nan_to_num(x / extended_phases_distances)


dsp.add_function(function=sh.bypass, weight=sh.inf(10, 300), inputs=[
    'phases_integration_times', 'cumulative_co2_emissions', 'phases_distances'
], outputs=[
    'extended_phases_integration_times', 'extended_cumulative_co2_emissions',
    'extended_phases_distances'
])


@sh.add_function(dsp, outputs=['cumulative_co2_emissions'])
def calculate_cumulative_co2_v1(phases_co2_emissions, phases_distances):
    """
    Calculates cumulative CO2 of cycle phases [CO2g].

    :param phases_co2_emissions:
        CO2 emission of cycle phases [CO2g/km].
    :type phases_co2_emissions: numpy.array

    :param phases_distances:
        Cycle phases distances [km].
    :type phases_distances: numpy.array

    :return:
        Cumulative CO2 of cycle phases [CO2g].
    :rtype: numpy.array
    """

    return phases_co2_emissions * phases_distances


def _define_rescaling_function(
        co2_emissions_model, cumulative_co2_emissions, phases_integration_times,
        times, rescaling_matrix):
    dx, it = np.append(np.diff(times), [0]), []
    for ii, jj in np.searchsorted(times, phases_integration_times):
        d = dx[ii:jj].copy()
        d[1:-1] = d[1:-1] + d[:-2]
        it.append((ii, jj, d[:, None] * rescaling_matrix[ii:jj, :] / 2))

    def _rescaling_function(params_initial_guess):
        co2_emissions = co2_emissions_model(params_initial_guess)[0]
        mtx = [np.sum(co2_emissions[i:j, None] * m, 0) for i, j, m in it]
        k_factors = np.linalg.lstsq(mtx, cumulative_co2_emissions, rcond=-1)[0]
        co2_emissions *= np.dot(rescaling_matrix, k_factors)
        return co2_emissions, k_factors

    return _rescaling_function


def _rescaling_matrix(
        phases_integration_times, times, velocities, stop_velocity):
    from scipy.interpolate import interp1d
    # noinspection PyProtectedMember
    d, eps = dfl.functions._rescaling_matrix, dfl.EPS
    a, b = np.array([-1, 1]) * d.a / 2, d.b
    pit = np.array(phases_integration_times)
    mean = np.mean(pit, 1)
    points = np.zeros((len(phases_integration_times), 4), float)
    points[0, 0], points[-1, 3] = -np.inf, np.inf
    points[1:, 0] = (pit[1:, 0] - mean[:-1]) * (1 - b) + mean[:-1] - eps
    points[:, 1:3] = np.column_stack((mean,) * 2) + np.diff(pit, axis=1) * a
    points[:-1, 3] = (mean[1:] - pit[:-1, 1]) * b + pit[:-1, 1]

    r, y = [], (0, 1, 1, 0)
    for x in points:
        r.append(interp1d(x, y, bounds_error=False, fill_value=0)(times))
    r = np.column_stack(r)
    r[np.isnan(r)] = 1
    b = np.asarray(velocities <= stop_velocity, int)
    # noinspection PyTypeChecker
    it = np.split(range(b.size), np.where(np.diff(b) != 0)[0] + 1)[1 - b[0]::2]
    for x in it:
        i = list(itertools.chain(*(np.where(v > 0)[0] for v in r[x])))
        r[x] = 0
        r[x, int(np.median(i))] = 1
    return r / np.sum(r, 1)[:, None]


def _rescaling_score(times, rescaling_matrix, k):
    x = np.dot(rescaling_matrix, k)
    m = np.trapz(x, times) / (times[-1] - times[0])
    std = np.sqrt(np.trapz((x - m) ** 2, times) / (times[-1] - times[0]))
    return m, std


@sh.add_function(dsp, weight=5, outputs=[
    'identified_co2_emissions', 'co2_rescaling_scores', 'co2_params_identified'
])
def identify_co2_emissions(
        co2_emissions_model, co2_params_initial_guess, times,
        extended_phases_integration_times, extended_cumulative_co2_emissions,
        engine_temperatures, is_cycle_hot, velocities, stop_velocity):
    """
    Identifies instantaneous CO2 emission vector [CO2g/s].

    :param co2_emissions_model:
        CO2 emissions model (co2_emissions = models(params)).
    :type co2_emissions_model: callable

    :param co2_params_initial_guess:
        Initial guess of co2 emission model params.
    :type co2_params_initial_guess: dict

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param extended_phases_integration_times:
        Extended cycle phases integration times [s].
    :type extended_phases_integration_times: tuple

    :param extended_cumulative_co2_emissions:
        Extended cumulative CO2 of cycle phases [CO2g].
    :type extended_cumulative_co2_emissions: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool

    :param velocities:
        Velocity vector [km/h].
    :type velocities: numpy.array

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        The instantaneous CO2 emission vector [CO2g/s], rescaling scores 
        (i.e., mean, std, and number of perturbations) [-], and the identified 
        initial guess of co2 emission model params.
    :rtype: numpy.array, tuple[float], dict
    """

    p = co2_params_initial_guess
    rescaling_matrix = _rescaling_matrix(
        extended_phases_integration_times, times, velocities, stop_velocity
    )
    rescale = _define_rescaling_function(
        co2_emissions_model, extended_cumulative_co2_emissions,
        extended_phases_integration_times, times, rescaling_matrix
    )

    d = dfl.functions.identify_co2_emissions
    n, (co2, k0) = 0, rescale(p)

    if d.enable_first_step or d.enable_second_step or d.enable_third_step:
        calibrate = functools.partial(
            calibrate_co2_params, is_cycle_hot, engine_temperatures,
            co2_emissions_model, _1st_step=d.enable_first_step,
            _2nd_step=d.enable_second_step, _3rd_step=d.enable_third_step,
        )
        xatol = d.xatol
        for n in range(d.n_perturbations):
            p = calibrate(co2, p)[0]
            co2, k1 = rescale(p)
            if np.max(np.abs(k1 - k0)) <= xatol:
                k0 = k1
                break
            k0 = k1

    return co2, _rescaling_score(times, rescaling_matrix, k0) + (n,), p


@sh.add_function(dsp, outputs=[
    'identified_co2_emissions', 'co2_rescaling_scores', 'co2_params_identified'
])
def fake_identification_co2_emissions(co2_emissions, co2_params_initial_guess):
    """
    Identifies instantaneous CO2 emission vector [CO2g/s].

    :param co2_emissions:
        CO2 instantaneous emissions vector [CO2g/s].
    :type co2_emissions: numpy.array

    :param co2_params_initial_guess:
        Initial guess of co2 emission model params.
    :type co2_params_initial_guess: dict
    
    :return:
        The instantaneous CO2 emission vector [CO2g/s], rescaling scores 
        (i.e., mean, std, and number of perturbations) [-], and the identified 
        initial guess of co2 emission model params.
    :rtype: numpy.array, tuple[float], dict
    """
    return co2_emissions, (1.0, 0, 0), co2_params_initial_guess


def _define_co2_error(co2_emissions_model, co2_emissions):
    def _error_func(params, sub_values=None):
        x = co2_emissions if sub_values is None else co2_emissions[sub_values]
        y = co2_emissions_model(params, sub_values=sub_values)[0]
        return co2_utl.mae(x, y)

    return _error_func


def _set_attr(params, data, default=False, attr='vary'):
    """
    Set attribute to CO2 emission model parameters.

    :param params:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).
    :type params: lmfit.Parameters

    :param data:
        Parameter ids to be set or key/value to set.
    :type data: iterable | dict

    :param default:
        Default value to set if a list of parameters ids is provided.
    :type default: bool | float

    :param attr:
        Parameter attribute to set.
    :type attr: str

    :return:
        CO2 emission model parameters.
    :rtype: lmfit.Parameters
    """
    import lmfit
    if not isinstance(data, dict):
        data = dict.fromkeys(data, default)

    d = {'min', 'max', 'value', 'vary', 'expr'} - {attr}

    for k, v in data.items():
        p = params[k]
        s = {i: getattr(p, i) for i in d}
        s[attr] = v
        p.set(**s)

        if lmfit.parameter.isclose(p.min, p.max, atol=1e-13, rtol=1e-13):
            p.set(value=np.mean((p.max, p.min)), min=None, max=None, vary=False)

    return params


def _identify_cold_phase(p, is_cycle_hot, engine_temperatures):
    cold = np.zeros_like(engine_temperatures, dtype=bool)
    if not is_cycle_hot:
        i = co2_utl.argmax(engine_temperatures >= p['trg'].value)
        cold[:i] = True
    return cold


def _calibrate_model_params(
        error_function, params, *args, method='nelder', **kws):
    """
    Calibrates the model params minimising the error_function.

    :param error_function:
        Model error function.
    :type error_function: callable

    :param params:
        Initial guess of model params.

        If not specified a brute force is used to identify the best initial
        guess with in the bounds.
    :type params: dict, optional

    :param method:
        Name of the fitting method to use.
    :type method: str, optional

    :return:
        Calibrated model params.
    :rtype: dict
    """

    if not any(p.vary for p in params.values()):
        return params, True

    import lmfit

    if callable(error_function):
        _error_f = error_function
    else:
        def _error_f(p, *a, **k):
            return sum(f(p, *a, **k) for f in error_function)

    min_e_and_p = [np.inf, copy.deepcopy(params)]

    def _err_func(p, *a, **kwargs):
        r = np.float32(_error_f(p, *a, **kwargs))

        if r < min_e_and_p[0]:
            min_e_and_p[0], min_e_and_p[1] = (r, copy.deepcopy(p))

        return r

    # See #7: Neither BFGS nor SLSQP fix "solution families".
    # leastsq: Improper input: N=6 must not exceed M=1.
    # nelder is stable (297 runs, 14 vehicles) [average time 181s/14 vehicles].
    # lbfgsb is unstable (2 runs, 4 vehicles) [average time 23s/4 vehicles].
    # cg is stable (20 runs, 4 vehicles) [average time 37s/4 vehicles].
    # newton: Jacobian is required for Newton-CG method
    # cobyla is unstable (8 runs, 4 vehicles) [average time 16s/4 vehicles].
    # tnc is unstable (6 runs, 4 vehicles) [average time 23s/4 vehicles].
    # dogleg: Jacobian is required for dogleg minimization.
    # slsqp is unstable (4 runs, 4 vehicles) [average time 18s/4 vehicles].
    # differential_evolution is unstable (1 runs, 4 vehicles)
    # [average time 270s/4 vehicles].
    res = lmfit.minimize(
        _err_func, params, args=args, kws=kws, method=method, nan_policy='omit'
    )

    # noinspection PyUnresolvedReferences
    return (res.params if res.success else min_e_and_p[1]), res.success


# noinspection PyUnresolvedReferences
@sh.add_function(dsp, outputs=['co2_params_calibrated', 'calibration_status'])
def calibrate_co2_params(
        is_cycle_hot, engine_temperatures, co2_emissions_model,
        identified_co2_emissions, co2_params_identified,
        _1st_step=dfl.functions.calibrate_co2_params.enable_first_step,
        _2nd_step=dfl.functions.calibrate_co2_params.enable_second_step,
        _3rd_step=dfl.functions.calibrate_co2_params.enable_third_step):
    """
    Calibrates the CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg
    ).

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param co2_emissions_model:
        CO2 emissions model (co2_emissions = models(params)).
    :type co2_emissions_model: callable

    :param identified_co2_emissions:
        CO2 instantaneous emissions vector [CO2g/s].
    :type identified_co2_emissions: numpy.array

    :param co2_params_identified:
        Identified initial guess of co2 emission model params.
    :type co2_params_identified: Parameters

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool

    :param _1st_step:
        Enable first step in the co2_params calibration? [-]
    :type _1st_step: bool

    :param _2nd_step:
        Enable second step in the co2_params calibration? [-]
    :type _2nd_step: bool

    :param _3rd_step:
        Enable third step in the co2_params calibration? [-]
    :type _3rd_step: bool

    :return:
        Calibrated CO2 emission model parameters (a2, b2, a, b, c, l, l2, t,
        trg) and their calibration statuses.
    :rtype: (lmfit.Parameters, list)
    """
    err = _define_co2_error(co2_emissions_model, identified_co2_emissions)

    # Safety measure to not modify the initial guess.
    p = copy.deepcopy(co2_params_identified)

    # Identify cold and hot phases.
    cold = _identify_cold_phase(p, is_cycle_hot, engine_temperatures)
    hot = ~cold

    # Definition of thermal and willans parameters.
    thermal_p = {'t0', 't1', 'dt', 'trg'}
    willans_p = {'a2', 'b2', 'a', 'b', 'c', 'l', 'l2'}

    # Identification of all parameters that can vary.
    pvary = {k for k, v in p.items() if v.vary}

    # Zero step: Initialization of the statuses.
    statuses = [(True, copy.deepcopy(p))]

    # Definition of the optimization function.
    def _op(par, params2optimize, **kws):
        fixp = pvary - params2optimize
        _set_attr(par, fixp)
        if pvary - fixp:
            par, s = _calibrate_model_params(err, par, **kws)
        else:
            s = True
        statuses.append((s, copy.deepcopy(par)))
        _set_attr(par, pvary, True)
        return par

    # First step: Calibration of willans parameters using the hot phase.
    p = _op(p, _1st_step and hot.any() and willans_p or set(), sub_values=hot)

    # Second step: Calibration of thermal parameters using the cold phase.
    if not cold.any():
        # When the cycle has not cold phases, thermal parameters have no effect.
        # The third step will modify arbitrarily this parameters.
        # Hence, to avoid erroneous results, thermal parameters are fixed to
        # zero because they cannot be identified.
        _set_attr(p, thermal_p)
        _set_attr(p, pvary.intersection(('t1', 'dt')), 0, 'value')
        pvary -= thermal_p

    p = _op(p, _2nd_step and cold.any() and thermal_p or set(), sub_values=cold)

    # Third step: Calibration of all parameters.
    p = _op(p, _3rd_step and pvary or set())

    return p, statuses


@sh.add_function(
    dsp, outputs=['co2_params_calibrated', 'calibration_status'],
    input_domain=functools.partial(_missing_co2_params, _not=True)
)
def define_co2_params_calibrated(co2_params):
    """
    Defines the calibrated co2_params if all co2_params are given.

    :param co2_params:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).
    :type co2_params: dict | lmfit.Parameters

    :return:
        Calibrated CO2 emission model parameters (a2, b2, a, b, c, l, l2, t,
        trg) and their calibration statuses.
    :rtype: (lmfit.Parameters, list)
    """
    import lmfit
    if isinstance(co2_params, lmfit.Parameters):
        p = co2_params
    else:
        p = lmfit.Parameters()
        for k, v in co2_params.items():
            p.add(k, value=v, vary=False)

    success = [(None, copy.deepcopy(p))] * 4

    return p, success


@sh.add_function(dsp, outputs=[
    'co2_emissions', 'active_cylinders', 'active_variable_valves',
    'active_lean_burns', 'active_exhausted_gas_recirculations'
])
def predict_co2_emissions(co2_emissions_model, co2_params_calibrated):
    """
    Predicts CO2 instantaneous emissions vector [CO2g/s].

    :param co2_emissions_model:
        CO2 emissions model (co2_emissions = models(params)).
    :type co2_emissions_model: callable

    :param co2_params_calibrated:
        CO2 emission model parameters (a2, b2, a, b, c, l, l2, t, trg).

        The missing parameters are set equal to zero.
    :type co2_params_calibrated: lmfit.Parameters

    :return:
        CO2 instantaneous emissions vector [CO2g/s].
    :rtype: numpy.array
    """

    return co2_emissions_model(co2_params_calibrated)


dsp.add_data('co2_params', dfl.values.co2_params.copy())


@sh.add_function(dsp, outputs=['fuel_consumptions'])
def calculate_fuel_consumptions(co2_emissions, fuel_carbon_content):
    """
    Calculates the instantaneous fuel consumption vector [g/s].

    :param co2_emissions:
        CO2 instantaneous emissions vector [CO2g/s].
    :type co2_emissions: numpy.array

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :return:
        The instantaneous fuel consumption vector [g/s].
    :rtype: numpy.array
    """
    return co2_emissions / fuel_carbon_content


@sh.add_function(dsp, outputs=['fuel_consumptions'])
def calculate_fuel_consumptions(fuel_consumptions_liters, fuel_density):
    """
    Calculates the instantaneous fuel consumption vector [g/s].

    :param fuel_consumptions_liters:
        The instantaneous fuel consumption vector [L/h].
    :type fuel_consumptions_liters: numpy.array

    :param fuel_density:
        Fuel density [g/l].
    :type fuel_density: float

    :return:
        The instantaneous fuel consumption vector [g/s].
    :rtype: numpy.array
    """
    return fuel_consumptions_liters * (fuel_density / 3600.0)


@sh.add_function(dsp, outputs=['co2_emissions'])
def calculate_co2_emissions(fuel_consumptions, fuel_carbon_content):
    """
    Calculates instantaneous CO2 emission vector [CO2g/s].

    :param fuel_consumptions:
        The instantaneous fuel consumption vector [g/s].
    :type fuel_consumptions: numpy.array

    :param fuel_carbon_content:
        Fuel carbon content [CO2g/g].
    :type fuel_carbon_content: float

    :return:
        CO2 instantaneous emissions vector [CO2g/s].
    :rtype: numpy.array
    """
    return fuel_consumptions * fuel_carbon_content


@sh.add_function(dsp, outputs=['fuel_consumptions_liters'])
def calculate_fuel_consumptions_liters(fuel_consumptions, fuel_density):
    """
    Calculates the instantaneous fuel consumption vector [g/s].

    :param fuel_consumptions_liters:
        The instantaneous fuel consumption vector [L/h].
    :type fuel_consumptions_liters: numpy.array

    :param fuel_density:
        Fuel density [g/l].
    :type fuel_density: float

    :return:
        The instantaneous fuel consumption vector [g/s].
    :rtype: numpy.array
    """
    return fuel_consumptions / (fuel_density / 3600.0)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the electric motor in P4 position.
"""
import numpy as np
import schedula as sh
import co2mpas.utils as co2_utl

dsp = sh.BlueDispatcher(
    name='Motor P4',
    description='Models the motor P4 (motor downstream the final drive).'
)


@sh.add_function(
    dsp, inputs=['motor_p4_front_powers'],
    outputs=['motor_p4_front_maximum_power']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_powers'],
    outputs=['motor_p4_rear_maximum_power']
)
def identify_motor_p4_maximum_power(motor_p4_powers):
    """
    Identify the maximum power of motor P4 [kW].

    :param motor_p4_powers:
        Power at motor P4 [kW].
    :type motor_p4_powers: numpy.array

    :return:
        Maximum power of motor P4 [kW].
    :rtype: float
    """
    return float(np.max(np.abs(motor_p4_powers)))


@sh.add_function(
    dsp, inputs=['motor_p4_front_torques'],
    outputs=['motor_p4_front_maximum_torque']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_torques'],
    outputs=['motor_p4_rear_maximum_torque']
)
def identify_motor_p4_maximum_torque(motor_p4_torques):
    """
    Identify the maximum torque of motor P4 [N*m].

    :param motor_p4_torques:
        Torque at motor P4 [N*m].
    :type motor_p4_torques: numpy.array

    :return:
        Maximum torque of motor P4 [N*m].
    :rtype: float
    """
    return float(np.max(np.abs(motor_p4_torques)))


@sh.add_function(
    dsp, inputs=['motor_p4_front_rated_speed', 'motor_p4_front_maximum_torque'],
    outputs=['motor_p4_front_maximum_power']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_rated_speed', 'motor_p4_rear_maximum_torque'],
    outputs=['motor_p4_rear_maximum_power']
)
def calculate_motor_p4_maximum_power(
        motor_p4_rated_speed, motor_p4_maximum_torque):
    """
    Calculate the maximum power of motor P4 [kW].

    :param motor_p4_rated_speed:
        Rated speed of motor P4 [RPM].
    :type motor_p4_rated_speed: float

    :param motor_p4_maximum_torque:
        Maximum torque of motor P4 [N*m].
    :type motor_p4_maximum_torque: float

    :return:
        Maximum power of motor P4 [kW].
    :rtype: float
    """
    from ...wheels import calculate_wheel_powers as func
    return func(motor_p4_maximum_torque, motor_p4_rated_speed)


@sh.add_function(
    dsp, outputs=['motor_p4_front_rated_speed'],
    inputs=['motor_p4_front_maximum_power', 'motor_p4_front_maximum_torque']
)
@sh.add_function(
    dsp, outputs=['motor_p4_rear_rated_speed'],
    inputs=['motor_p4_rear_maximum_power', 'motor_p4_rear_maximum_torque']
)
def calculate_motor_p4_rated_speed(
        motor_p4_maximum_power, motor_p4_maximum_torque):
    """
    Calculate the rated speed of motor P4 [RPM].

    :param motor_p4_maximum_power:
        Maximum power of motor P4 [kW].
    :type motor_p4_maximum_power: float

    :param motor_p4_maximum_torque:
        Maximum torque of motor P4 [N*m].
    :type motor_p4_maximum_torque: float

    :return:
        Rated speed of motor P4 [RPM].
    :rtype: float
    """
    from ...wheels import calculate_wheel_torques as func
    return func(motor_p4_maximum_power, motor_p4_maximum_torque)


@sh.add_function(
    dsp, outputs=['motor_p4_front_maximum_torque'],
    inputs=['motor_p4_front_maximum_power', 'motor_p4_front_rated_speed']
)
@sh.add_function(
    dsp, outputs=['motor_p4_rear_maximum_torque'],
    inputs=['motor_p4_rear_maximum_power', 'motor_p4_rear_rated_speed']
)
def calculate_motor_p4_maximum_torque(
        motor_p4_maximum_power, motor_p4_rated_speed):
    """
    Calculate the maximum torque of motor P4 [N*m].

    :param motor_p4_maximum_power:
        Maximum power of motor P4 [kW].
    :type motor_p4_maximum_power: float

    :param motor_p4_rated_speed:
        Rated speed of motor P4 [RPM].
    :type motor_p4_rated_speed: float

    :return:
        Maximum torque of motor P4 [N*m].
    :rtype: float
    """
    from ...wheels import calculate_wheel_torques as func
    return func(motor_p4_maximum_power, motor_p4_rated_speed)


@sh.add_function(
    dsp, inputs=['motor_p4_front_maximum_power', 'motor_p4_front_rated_speed'],
    outputs=['motor_p4_front_maximum_power_function']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_maximum_power', 'motor_p4_rear_rated_speed'],
    outputs=['motor_p4_rear_maximum_power_function']
)
def define_motor_p4_maximum_power_function(
        motor_p4_maximum_power, motor_p4_rated_speed):
    """
    Define the maximum power function of motor P4.

    :param motor_p4_maximum_power:
        Maximum power of motor P4 [kW].
    :type motor_p4_maximum_power: float

    :param motor_p4_rated_speed:
        Rated speed of motor P4 [RPM].
    :type motor_p4_rated_speed: float

    :return:
        Maximum power function of motor P4.
    :rtype: function
    """
    m = 0
    if motor_p4_rated_speed:
        m = motor_p4_maximum_power / motor_p4_rated_speed

    def _maximum_power_function(speeds):
        return np.minimum(motor_p4_maximum_power, speeds * m)

    return _maximum_power_function


@sh.add_function(
    dsp, outputs=['motor_p4_front_maximum_powers'],
    inputs=['motor_p4_front_speeds', 'motor_p4_front_maximum_power_function']
)
@sh.add_function(
    dsp, outputs=['motor_p4_rear_maximum_powers'],
    inputs=['motor_p4_rear_speeds', 'motor_p4_rear_maximum_power_function']
)
def calculate_motor_p4_maximum_powers(
        motor_p4_speeds, motor_p4_maximum_power_function):
    """
    Calculate the maximum power vector of motor P4 [kW].

    :param motor_p4_speeds:
        Rotating speed of motor P4 [RPM].
    :type motor_p4_speeds: numpy.array | float

    :param motor_p4_maximum_power_function:
        Maximum power function of motor P4.
    :type motor_p4_maximum_power_function: function

    :return:
        Maximum power vector of motor P4 [kW].
    :rtype: numpy.array | float
    """
    return motor_p4_maximum_power_function(motor_p4_speeds)


@sh.add_function(
    dsp, inputs=['wheel_speeds', 'motor_p4_front_speeds'],
    outputs=['motor_p4_front_speed_ratio']
)
@sh.add_function(
    dsp, inputs=['wheel_speeds', 'motor_p4_rear_speeds'],
    outputs=['motor_p4_rear_speed_ratio']
)
def identify_motor_p4_speed_ratio(wheel_speeds, motor_p4_speeds):
    """
    Identifies motor P4 speed ratio.

    :param wheel_speeds:
        Rotating speed of the wheel [RPM].
    :type wheel_speeds: numpy.array | float

    :param motor_p4_speeds:
        Rotating speed of motor P4 [RPM].
    :type motor_p4_speeds: numpy.array | float

    :return:
        Motor P4 speed ratio [-].
    :rtype: float
    """
    b = wheel_speeds > 0
    return co2_utl.reject_outliers(motor_p4_speeds[b] / wheel_speeds[b])[0]


dsp.add_data('motor_p4_front_speed_ratio', 1, sh.inf(10, 1))
dsp.add_data('motor_p4_rear_speed_ratio', 1, sh.inf(10, 1))


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['motor_p4_front_speeds'],
    inputs=['wheel_speeds', 'motor_p4_front_speed_ratio']
)
@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['motor_p4_rear_speeds'],
    inputs=['wheel_speeds', 'motor_p4_rear_speed_ratio']
)
def calculate_motor_p4_speeds(wheel_speeds, motor_p4_speed_ratio=1):
    """
    Calculates rotating speed of motor P4 [RPM].

    :param wheel_speeds:
        Rotating speed of the wheel [RPM].
    :type wheel_speeds: numpy.array | float

    :param motor_p4_speed_ratio:
        Ratio between motor P4 speed and wheel speed [-].
    :type motor_p4_speed_ratio: float

    :return:
        Rotating speed of motor P4 [RPM].
    :rtype: numpy.array | float
    """
    return wheel_speeds * motor_p4_speed_ratio


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['wheel_speeds'],
    inputs=['motor_p4_front_speeds', 'motor_p4_front_speed_ratio']
)
@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['wheel_speeds'],
    inputs=['motor_p4_rear_speeds', 'motor_p4_rear_speed_ratio']
)
def calculate_wheel_speeds(motor_p4_speeds, motor_p4_speed_ratio=1):
    """
    Calculates rotating speed of the wheels [RPM].

    :param motor_p4_speeds:
        Rotating speed of motor P4 [RPM].
    :type motor_p4_speeds: numpy.array | float

    :param motor_p4_speed_ratio:
        Ratio between motor P4 speed and wheel speed [-].
    :type motor_p4_speed_ratio: float

    :return:
        Rotating speed of the wheel [RPM].
    :rtype: numpy.array | float
    """
    return motor_p4_speeds / motor_p4_speed_ratio


@sh.add_function(
    dsp, inputs=['motor_p4_front_powers', 'motor_p4_front_speeds'],
    outputs=['motor_p4_front_torques']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_powers', 'motor_p4_rear_speeds'],
    outputs=['motor_p4_rear_torques']
)
def calculate_motor_p4_torques(motor_p4_powers, motor_p4_speeds):
    """
    Calculates torque at motor P4 [N*m].

    :param motor_p4_powers:
        Power at motor P4 [kW].
    :type motor_p4_powers: numpy.array | float

    :param motor_p4_speeds:
        Rotating speed of motor P4 [RPM].
    :type motor_p4_speeds: numpy.array | float

    :return:
        Torque at motor P4 [N*m].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques
    return calculate_wheel_torques(motor_p4_powers, motor_p4_speeds)


@sh.add_function(
    dsp, inputs=['motor_p4_front_torques', 'motor_p4_front_speeds'],
    outputs=['motor_p4_front_powers']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_torques', 'motor_p4_rear_speeds'],
    outputs=['motor_p4_rear_powers']
)
def calculate_motor_p4_powers(motor_p4_torques, motor_p4_speeds):
    """
    Calculates power at motor P4 [kW].

    :param motor_p4_torques:
        Torque at motor P4 [N*m].
    :type motor_p4_torques: numpy.array | float

    :param motor_p4_speeds:
        Rotating speed of motor P4 [RPM].
    :type motor_p4_speeds: numpy.array | float

    :return:
        Power at motor P4 [kW].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_powers
    return calculate_wheel_powers(motor_p4_torques, motor_p4_speeds)


@sh.add_function(
    dsp, inputs=['motor_p4_front_powers', 'motor_p4_front_torques'],
    outputs=['motor_p4_front_speeds']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_powers', 'motor_p4_rear_torques'],
    outputs=['motor_p4_rear_speeds']
)
def calculate_motor_p4_speeds_v1(motor_p4_powers, motor_p4_torques):
    """
    Calculates rotating speed of motor P4 [RPM].

    :param motor_p4_powers:
        Power at motor P4 [kW].
    :type motor_p4_powers: numpy.array | float

    :param motor_p4_torques:
        Torque at motor P4 [N*m].
    :type motor_p4_torques: numpy.array | float

    :return:
        Rotating speed of motor P4 [RPM].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques
    return calculate_wheel_torques(motor_p4_powers, motor_p4_torques)


dsp.add_data('motor_p4_front_efficiency', 0.9)
dsp.add_data('motor_p4_rear_efficiency', 0.9)


@sh.add_function(
    dsp, inputs=['motor_p4_front_powers', 'motor_p4_front_efficiency'],
    outputs=['motor_p4_front_electric_powers']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_powers', 'motor_p4_rear_efficiency'],
    outputs=['motor_p4_rear_electric_powers']
)
def calculate_motor_p4_electric_powers(motor_p4_powers, motor_p4_efficiency):
    """
    Calculates motor P4 electric power [kW].

    :param motor_p4_powers:
        Power at motor P4 [kW].
    :type motor_p4_powers: numpy.array | float

    :param motor_p4_efficiency:
        Motor P4 efficiency [-].
    :type motor_p4_efficiency: float

    :return:
        Electric power of motor P4 [kW].
    :rtype: numpy.array | float
    """
    p = motor_p4_powers
    return p * np.where(p >= 0, 1 / motor_p4_efficiency, motor_p4_efficiency)


@sh.add_function(
    dsp, inputs=['motor_p4_front_electric_powers', 'motor_p4_front_efficiency'],
    outputs=['motor_p4_front_powers'], weight=5
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_electric_powers', 'motor_p4_rear_efficiency'],
    outputs=['motor_p4_rear_powers'], weight=5
)
def calculate_motor_p4_powers_v1(motor_p4_electric_powers, motor_p4_efficiency):
    """
    Calculate motor P4 power from electric power and electric power losses [kW].

    :param motor_p4_electric_powers:
        Electric power of motor P4 [kW].
    :type motor_p4_electric_powers: numpy.array | float

    :param motor_p4_efficiency:
        Motor P4 efficiency [-].
    :type motor_p4_efficiency: float

    :return:
        Power at motor P4 [kW].
    :rtype: numpy.array | float
    """
    p = motor_p4_electric_powers
    return p * np.where(p < 0, 1 / motor_p4_efficiency, motor_p4_efficiency)


dsp.add_data('has_motor_p4_front', False, sh.inf(10, 3))
dsp.add_data('has_motor_p4_rear', False, sh.inf(10, 3))


@sh.add_function(
    dsp, inputs=['motor_p4_front_maximum_power'], outputs=['has_motor_p4_front']
)
@sh.add_function(
    dsp, inputs=['motor_p4_rear_maximum_power'], outputs=['has_motor_p4_rear']
)
def identify_has_motor_p4(motor_p4_maximum_power):
    """
    Identify if the vehicle has a motor P4 [kW].

    :param motor_p4_maximum_power:
        Maximum power of motor P4 [kW].
    :type motor_p4_maximum_power: float

    :return:
        Has the vehicle a motor in P4?
    :rtype: bool
    """
    return not np.isclose(motor_p4_maximum_power, 0)


@sh.add_function(
    dsp, inputs=['times', 'has_motor_p4_front'],
    outputs=['motor_p4_front_powers']
)
@sh.add_function(
    dsp, inputs=['times', 'has_motor_p4_rear'], outputs=['motor_p4_rear_powers']
)
def default_motor_p4_powers(times, has_motor_p4):
    """
    Return zero power if the vehicle has not a motor P4 [kW].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param has_motor_p4:
        Has the vehicle a motor in P4?
    :type has_motor_p4: bool

    :return:
        Power at motor P4 [kW].
    :rtype: numpy.array
    """
    if not has_motor_p4:
        return np.zeros_like(times, float)
    return sh.NONE
